"""
Contains possible interactions with the Galaxy Datasets
"""

import logging
import os
import shlex
import warnings
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    overload,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from requests import Response

from bioblend import (
    CHUNK_SIZE,
    NotReady,
    TimeoutException,
    wait_on,
)
from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

log = logging.getLogger(__name__)

HdaLdda = Literal["hda", "ldda"]
TERMINAL_STATES = {"ok", "empty", "error", "discarded", "failed_metadata"}
# Non-terminal states are: 'new', 'upload', 'queued', 'running', 'paused', 'setting_metadata'


class DatasetClient(Client):
    gi: "GalaxyInstance"
    module = "datasets"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def show_dataset(self, dataset_id: str, hda_ldda: HdaLdda = "hda") -> Dict[str, Any]:
        """
        Get details about a given dataset. This can be a history or a library dataset.

        :type dataset_id: str
        :param dataset_id: Encoded dataset ID

        :type hda_ldda: str
        :param hda_ldda: Whether to show a history dataset ('hda' - the default) or library
                         dataset ('ldda').

        :rtype: dict
        :return: Information about the HDA or LDDA
        """
        params = {
            "hda_ldda": hda_ldda,
        }
        return self._get(id=dataset_id, params=params)

    def _initiate_download(
        self, dataset_id: str, stream_content: bool, require_ok_state: bool = True, maxwait: float = 12000
    ) -> Tuple[Dict[str, Any], str, Response]:
        dataset = self.wait_for_dataset(dataset_id, maxwait=maxwait, check=False)
        if not dataset["state"] == "ok":
            message = f"Dataset state is not 'ok'. Dataset id: {dataset_id}, current state: {dataset['state']}"
            if require_ok_state:
                raise DatasetStateException(message)
            else:
                warnings.warn(message, DatasetStateWarning, stacklevel=2)

        file_ext = dataset.get("file_ext")
        # Resort to 'data' when Galaxy returns an empty or temporary extension
        if not file_ext or file_ext == "auto" or file_ext == "_sniff_":
            file_ext = "data"
        # The preferred download URL is
        # '/api/histories/<history_id>/contents/<dataset_id>/display?to_ext=<dataset_ext>'
        # since the old URL:
        # '/dataset/<dataset_id>/display?to_ext=<dataset_ext>'
        # does not work when using REMOTE_USER with access disabled to
        # everything but /api without auth
        download_url = dataset["download_url"] + "?to_ext=" + file_ext
        url = f"{self.gi.base_url}{download_url}"

        r = self.gi.make_get_request(url, stream=stream_content)
        r.raise_for_status()
        return dataset, file_ext, r

    @overload
    def download_dataset(
        self,
        dataset_id: str,
        file_path: None = None,
        use_default_filename: bool = True,
        require_ok_state: bool = True,
        maxwait: float = 12000,
    ) -> bytes: ...

    @overload
    def download_dataset(
        self,
        dataset_id: str,
        file_path: str,
        use_default_filename: bool = True,
        require_ok_state: bool = True,
        maxwait: float = 12000,
    ) -> str: ...

    def download_dataset(
        self,
        dataset_id: str,
        file_path: Optional[str] = None,
        use_default_filename: bool = True,
        require_ok_state: bool = True,
        maxwait: float = 12000,
    ) -> Union[bytes, str]:
        """
        Download a dataset to file or in memory. If the dataset state is not
        'ok', a ``DatasetStateException`` will be thrown, unless ``require_ok_state=False``.

        :type dataset_id: str
        :param dataset_id: Encoded dataset ID

        :type file_path: str
        :param file_path: If this argument is provided, the dataset will be streamed to disk
                          at that path (should be a directory if ``use_default_filename=True``).
                          If the file_path argument is not provided, the dataset content is loaded into memory
                          and returned by the method (Memory consumption may be heavy as the entire file
                          will be in memory).

        :type use_default_filename: bool
        :param use_default_filename: If ``True``, the exported
                                 file will be saved as ``file_path/%s``,
                                 where ``%s`` is the dataset name.
                                 If ``False``, ``file_path`` is assumed to
                                 contain the full file path including the filename.

        :type require_ok_state: bool
        :param require_ok_state: If ``False``, datasets will be downloaded even if not in an 'ok' state,
                                 issuing a ``DatasetStateWarning`` rather than raising a ``DatasetStateException``.

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the dataset state to
          become terminal. After this time, a ``TimeoutException`` will be
          raised.

        :rtype: bytes or str
        :return: If a ``file_path`` argument is not provided, returns the file
          content. Otherwise returns the local path of the downloaded file.
        """
        dataset, file_ext, r = self._initiate_download(
            dataset_id, stream_content=file_path is not None, require_ok_state=require_ok_state, maxwait=maxwait
        )
        if file_path is None:
            if "content-length" in r.headers and len(r.content) != int(r.headers["content-length"]):
                log.warning(
                    "Transferred content size does not match content-length header (%s != %s)",
                    len(r.content),
                    r.headers["content-length"],
                )
            return r.content
        else:
            if use_default_filename:
                # Build a useable filename
                filename = dataset["name"] + "." + file_ext
                # Now try to get a better filename from the response headers
                # We expect tokens 'filename' '=' to be followed by the quoted filename
                if "content-disposition" in r.headers:
                    tokens = list(shlex.shlex(r.headers["content-disposition"], posix=True))
                    try:
                        header_filepath = tokens[tokens.index("filename") + 2]
                        filename = os.path.basename(header_filepath)
                    except (ValueError, IndexError):
                        pass
                file_local_path = os.path.join(file_path, filename)
            else:
                file_local_path = file_path

            with open(file_local_path, "wb") as fp:
                for chunk in r.iter_content(chunk_size=CHUNK_SIZE):
                    if chunk:
                        fp.write(chunk)

            # Return location file was saved to
            return file_local_path

    def get_datasets(
        self,
        limit: int = 500,
        offset: int = 0,
        name: Optional[str] = None,
        extension: Optional[Union[str, List[str]]] = None,
        state: Optional[Union[str, List[str]]] = None,
        visible: Optional[bool] = None,
        deleted: Optional[bool] = None,
        purged: Optional[bool] = None,
        tool_id: Optional[str] = None,
        tag: Optional[str] = None,
        history_id: Optional[str] = None,
        create_time_min: Optional[str] = None,
        create_time_max: Optional[str] = None,
        update_time_min: Optional[str] = None,
        update_time_max: Optional[str] = None,
        order: str = "create_time-dsc",
    ) -> List[Dict[str, Any]]:
        """
        Get the latest datasets, or select another subset by specifying optional
        arguments for filtering (e.g. a history ID).

        Since the number of datasets may be very large, ``limit`` and ``offset``
        parameters are required to specify the desired range.

        If the user is an admin, this will return datasets for all the users,
        otherwise only for the current user.

        :type limit: int
        :param limit: Maximum number of datasets to return.

        :type offset: int
        :param offset: Return datasets starting from this specified position.
          For example, if ``limit`` is set to 100 and ``offset`` to 200,
          datasets 200-299 will be returned.

        :type name: str
        :param name: Dataset name to filter on.

        :type extension: str or list of str
        :param extension: Dataset extension (or list of extensions) to filter on.

        :type state: str or list of str
        :param state: Dataset state (or list of states) to filter on.

        :type visible: bool
        :param visible: Optionally filter datasets by their ``visible`` attribute.

        :type deleted: bool
        :param deleted: Optionally filter datasets by their ``deleted`` attribute.

        :type purged: bool
        :param purged: Optionally filter datasets by their ``purged`` attribute.

        :type tool_id: str
        :param tool_id: Tool ID to filter on.

        :type tag: str
        :param tag: Dataset tag to filter on.

        :type history_id: str
        :param history_id: Encoded history ID to filter on.

        :type create_time_min: str
        :param create_time_min: Show only datasets created after the provided
          time and date, which should be formatted as ``YYYY-MM-DDTHH-MM-SS``.

        :type create_time_max: str
        :param create_time_max: Show only datasets created before the provided
          time and date, which should be formatted as ``YYYY-MM-DDTHH-MM-SS``.

        :type update_time_min: str
        :param update_time_min: Show only datasets last updated after the provided
          time and date, which should be formatted as ``YYYY-MM-DDTHH-MM-SS``.

        :type update_time_max: str
        :param update_time_max: Show only datasets last updated before the provided
          time and date, which should be formatted as ``YYYY-MM-DDTHH-MM-SS``.

        :type order: str
        :param order: One or more of the following attributes for ordering datasets:
          ``create_time`` (default), ``extension``, ``hid``, ``history_id``, ``name``,
          ``update_time``. Optionally, ``-asc`` or ``-dsc`` (default) can be appended
          for ascending and descending order respectively. Multiple attributes can be
          stacked as a comma-separated list of values, e.g. ``create_time-asc,hid-dsc``.

        :rtype: list
        :param: A list of datasets
        """
        params: Dict[str, Any] = {
            "limit": limit,
            "offset": offset,
            "order": order,
        }
        if history_id:
            params["history_id"] = history_id

        q: List[str] = []
        qv = []

        if name:
            q.append("name")
            qv.append(name)
        if state:
            op, val = self._param_to_filter(state)
            q.append(f"state-{op}")
            qv.append(val)
        if extension:
            op, val = self._param_to_filter(extension)
            q.append(f"extension-{op}")
            qv.append(val)
        if visible is not None:
            q.append("visible")
            qv.append(str(visible))
        if deleted is not None:
            q.append("deleted")
            qv.append(str(deleted))
        if purged is not None:
            q.append("purged")
            qv.append(str(purged))
        if tool_id is not None:
            q.append("tool_id")
            qv.append(str(tool_id))
        if tag is not None:
            q.append("tag")
            qv.append(str(tag))
        if create_time_min:
            q.append("create_time-ge")
            qv.append(create_time_min)
        if create_time_max:
            q.append("create_time-le")
            qv.append(create_time_max)
        if update_time_min:
            q.append("update_time-ge")
            qv.append(update_time_min)
        if update_time_max:
            q.append("update_time-le")
            qv.append(update_time_max)

        params["q"] = q
        params["qv"] = qv

        return self._get(params=params)

    def _param_to_filter(self, param: Union[str, List[str]]) -> Tuple[str, str]:
        if isinstance(param, str):
            return "eq", param
        if isinstance(param, list):
            if len(param) == 1:
                return "eq", param.pop()
            return "in", ",".join(param)
        raise Exception("Filter param is not of type ``str`` or ``list``")

    def publish_dataset(self, dataset_id: str, published: bool = False) -> Dict[str, Any]:
        """
        Make a dataset publicly available or private. For more fine-grained control (assigning different
        permissions to specific roles), use the ``update_permissions()`` method.

        :type dataset_id: str
        :param dataset_id: dataset ID

        :type published: bool
        :param published: Whether to make the dataset published (``True``) or private (``False``).

        :rtype: dict
        :return: Details of the updated dataset

        .. note::
          This method works only on Galaxy 19.05 or later.
        """
        payload: Dict[str, Any] = {"action": "remove_restrictions" if published else "make_private"}
        url = self._make_url(dataset_id) + "/permissions"
        return self.gi.datasets._put(url=url, payload=payload)

    def update_permissions(
        self,
        dataset_id: str,
        access_ids: Optional[list] = None,
        manage_ids: Optional[list] = None,
        modify_ids: Optional[list] = None,
    ) -> dict:
        """
        Set access, manage or modify permissions for a dataset to a list of roles.

        :type dataset_id: str
        :param dataset_id: dataset ID

        :type access_ids: list
        :param access_ids: role IDs which should have access permissions for the dataset.

        :type manage_ids: list
        :param manage_ids: role IDs which should have manage permissions for the dataset.

        :type modify_ids: list
        :param modify_ids: role IDs which should have modify permissions for the dataset.

        :rtype: dict
        :return: Current roles for all available permission types.

        .. note::
          This method works only on Galaxy 19.05 or later.
        """
        payload: Dict[str, Any] = {"action": "set_permissions"}
        if access_ids:
            payload["access"] = access_ids
        if manage_ids:
            payload["manage"] = manage_ids
        if modify_ids:
            payload["modify"] = modify_ids
        url = self._make_url(dataset_id) + "/permissions"
        return self.gi.datasets._put(url=url, payload=payload)

    def wait_for_dataset(
        self, dataset_id: str, maxwait: float = 12000, interval: float = 3, check: bool = True
    ) -> Dict[str, Any]:
        """
        Wait until a dataset is in a terminal state.

        :type dataset_id: str
        :param dataset_id: dataset ID

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the dataset state to
          become terminal. After this time, a ``TimeoutException`` will be raised.

        :type interval: float
        :param interval: Time (in seconds) to wait between 2 consecutive checks.

        :type check: bool
        :param check: Whether to check if the dataset terminal state is 'ok'.

        :rtype: dict
        :return: Details of the given dataset.
        """

        def check_and_get_dataset() -> Dict[str, Any]:
            dataset = self.show_dataset(dataset_id)
            state = dataset["state"]
            if state in TERMINAL_STATES:
                if check and state != "ok":
                    raise Exception(f"Dataset {dataset_id} is in terminal state {state}")
                return dataset
            raise NotReady(f"Dataset {dataset_id} is in non-terminal state {state}")

        return wait_on(check_and_get_dataset, maxwait=maxwait, interval=interval)


class DatasetStateException(Exception):
    pass


class DatasetStateWarning(UserWarning):
    pass


# Unused, just for backward compatibility
DatasetTimeoutException = TimeoutException



"""
Contains possible interactions with the Galaxy Forms
"""

from typing import (
    Any,
    Dict,
    List,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class FormsClient(Client):
    module = "forms"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_forms(self) -> List[Dict[str, Any]]:
        """
        Get the list of all forms.

        :rtype: list
        :return: Displays a collection (list) of forms.
          For example::

            [{'id': 'f2db41e1fa331b3e',
              'model_class': 'FormDefinition',
              'name': 'First form',
              'url': '/api/forms/f2db41e1fa331b3e'},
             {'id': 'ebfb8f50c6abde6d',
              'model_class': 'FormDefinition',
              'name': 'second form',
              'url': '/api/forms/ebfb8f50c6abde6d'}]
        """
        return self._get()

    def show_form(self, form_id: str) -> Dict[str, Any]:
        """
        Get details of a given form.

        :type form_id: str
        :param form_id: Encoded form ID

        :rtype: dict
        :return: A description of the given form.
          For example::

            {'desc': 'here it is ',
             'fields': [],
             'form_definition_current_id': 'f2db41e1fa331b3e',
             'id': 'f2db41e1fa331b3e',
             'layout': [],
             'model_class': 'FormDefinition',
             'name': 'First form',
             'url': '/api/forms/f2db41e1fa331b3e'}
        """
        return self._get(id=form_id)

    def create_form(self, form_xml_text: str) -> List[Dict[str, Any]]:
        """
        Create a new form.

        :type form_xml_text: str
        :param form_xml_text: Form xml to create a form on galaxy instance

        :rtype: list of dicts
        :return: List with a single dictionary describing the created form
        """
        payload = {
            "xml_text": form_xml_text,
        }
        return self._post(payload=payload)



from . import GalaxyTestBase


class TestGalaxyToolData(GalaxyTestBase.GalaxyTestBase):
    def test_get_data_tables(self):
        tables = self.gi.tool_data.get_data_tables()
        for table in tables:
            assert table["name"] is not None

    def test_show_data_table(self):
        tables = self.gi.tool_data.get_data_tables()
        table = self.gi.tool_data.show_data_table(tables[0]["name"])
        assert table["columns"] is not None
        assert table["fields"] is not None
        assert table["name"] is not None



"""
"""

import os
import shutil
import tarfile
import tempfile

import pytest

from bioblend import (
    ConnectionError,
    galaxy,
)
from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyHistories(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.default_history_name = "buildbot - automated test"
        self.history = self.gi.histories.create_history(name=self.default_history_name)

    def test_create_history(self):
        history_name = "another buildbot - automated test"
        new_history = self.gi.histories.create_history(name=history_name)
        assert new_history["id"] is not None
        assert new_history["name"] == history_name
        assert new_history["url"] is not None

    def test_update_history(self):
        new_name = "buildbot - automated test renamed"
        new_annotation = f"Annotation for {new_name}"
        new_tags = ["tag1", "tag2"]
        updated_hist = self.gi.histories.update_history(
            self.history["id"], name=new_name, annotation=new_annotation, tags=new_tags
        )
        if "id" not in updated_hist:
            updated_hist = self.gi.histories.show_history(self.history["id"])
        assert self.history["id"] == updated_hist["id"]
        assert updated_hist["name"] == new_name
        assert updated_hist["annotation"] == new_annotation
        assert updated_hist["tags"] == new_tags

    def test_publish_history(self):
        # Verify that searching for published histories does not return the test history
        published_histories = self.gi.histories.get_histories(published=True)
        assert not any(h["id"] == self.history["id"] for h in published_histories)
        updated_hist = self.gi.histories.update_history(self.history["id"], published=True)
        if "id" not in updated_hist:
            updated_hist = self.gi.histories.show_history(self.history["id"])
        assert self.history["id"] == updated_hist["id"]
        assert updated_hist["published"]
        # Verify that searching for published histories now returns the test history
        published_histories = self.gi.histories.get_histories(published=True)
        assert any(h["id"] == self.history["id"] for h in published_histories)
        # Verify that get_published_histories as an anonymous user also returns the test history
        anonymous_gi = galaxy.GalaxyInstance(url=self.gi.base_url, key=None)
        published_histories = anonymous_gi.histories.get_published_histories()
        assert any(h["id"] == self.history["id"] for h in published_histories)
        history_from_slug = anonymous_gi.histories.get_published_histories(slug=updated_hist["slug"])
        assert len(history_from_slug) == 1
        assert self.history["id"] == history_from_slug[0]["id"]

    def test_get_histories(self):
        # Make sure there's at least one value - the one we created
        all_histories = self.gi.histories.get_histories()
        assert len(all_histories) > 0

        # Check whether id is present, when searched by name
        histories = self.gi.histories.get_histories(name=self.default_history_name)
        assert len([h for h in histories if h["id"] == self.history["id"]]) == 1

        # Test for time filters. We expect all histories to have been created an updated within the last few minutes
        new_histories = self.gi.histories.get_histories(
            create_time_min="2023-07-04T11:00:01", update_time_min="2023-08-06T11:00:05"
        )
        assert len(new_histories) == len(all_histories)

        old_histories = self.gi.histories.get_histories(
            create_time_max="2023-07-04T11:00:01", update_time_max="2023-08-06T11:00:05"
        )
        assert len(old_histories) == 0

        # Test detailed view: check for presence of "size" field
        histories_detailed = self.gi.histories.get_histories(view="detailed")
        assert "size" in histories_detailed[0]

        # Test keys: check that fields requested are returned
        histories_with_keys = self.gi.histories.get_histories(keys=["id", "user_id", "size"])
        assert set(histories_with_keys[0]) >= {"id", "user_id", "size"}

        # TODO: check whether deleted history is returned correctly
        # At the moment, get_histories() returns only not-deleted histories
        # and get_histories(deleted=True) returns only deleted histories,
        # so they are not comparable.
        # In the future, according to https://trello.com/c/MoilsmVv/1673-api-incoherent-and-buggy-indexing-of-deleted-entities ,
        # get_histories() will return both not-deleted and deleted histories
        # and we can uncomment the following test.
        # deleted_history = self.gi.histories.get_histories(deleted=True)
        # assert len(all_histories) >= len(deleted_history)

    @test_util.skip_unless_galaxy("release_20.01")
    def test_other_users_histories(self):
        username = test_util.random_string()
        user_id = self.gi.users.create_local_user(username, f"{username}@example.org", test_util.random_string(20))[
            "id"
        ]
        user_api_key = self.gi.users.create_user_apikey(user_id)
        user_gi = galaxy.GalaxyInstance(url=self.gi.base_url, key=user_api_key)
        # Normal users cannot use the `all` parameter
        with pytest.raises(ConnectionError):
            other_user_histories = user_gi.histories.get_histories(all=True)
        user_history_id = user_gi.histories.create_history(name=f"History for {username}")["id"]
        # Get all users' histories from an admin account
        other_user_histories = self.gi.histories.get_histories(all=True)
        assert user_history_id in [h["id"] for h in other_user_histories]

    def test_show_history(self):
        history_data = self.gi.histories.show_history(self.history["id"])
        assert self.history["id"] == history_data["id"]
        assert self.history["name"] == history_data["name"]
        assert "new" == history_data["state"]

    def test_show_history_with_contents(self):
        history_id = self.history["id"]
        contents = self.gi.histories.show_history(history_id, contents=True)
        # Empty history has no datasets, content length should be 0
        assert len(contents) == 0
        self._test_dataset(history_id)
        contents = self.gi.histories.show_history(history_id, contents=True)
        # history has 1 dataset, content length should be 1
        assert len(contents) == 1
        contents = self.gi.histories.show_history(history_id, contents=True, types=["dataset"])
        # filtering for dataset, content length should still be 1
        assert len(contents) == 1
        contents = self.gi.histories.show_history(history_id, contents=True, types=["dataset_collection"])
        # filtering for dataset collection but there's no collection in the history
        assert len(contents) == 0
        contents = self.gi.histories.show_history(history_id, contents=True, types=["dataset", "dataset_collection"])
        assert len(contents) == 1

    def test_create_history_tag(self):
        new_tag = "tag1"
        self.gi.histories.create_history_tag(self.history["id"], new_tag)
        updated_hist = self.gi.histories.show_history(self.history["id"])
        assert self.history["id"] == updated_hist["id"]
        assert new_tag in updated_hist["tags"]

    def test_show_dataset(self):
        history_id = self.history["id"]
        dataset1_id = self._test_dataset(history_id)
        dataset = self.gi.histories.show_dataset(history_id, dataset1_id)
        for key in ["name", "hid", "id", "deleted", "history_id", "visible"]:
            assert key in dataset
        assert dataset["history_id"] == history_id
        assert dataset["hid"] == 1
        assert dataset["id"] == dataset1_id
        assert not dataset["deleted"]
        assert dataset["visible"]

    @test_util.skip_unless_galaxy("release_22.01")
    def test_show_dataset_provenance(self) -> None:
        MINIMAL_PROV_KEYS = ("id", "uuid")
        OTHER_PROV_KEYS = ("job_id", "parameters", "stderr", "stdout", "tool_id")
        ALL_PROV_KEYS = MINIMAL_PROV_KEYS + OTHER_PROV_KEYS
        history_id = self.history["id"]
        dataset1_id = self._test_dataset(history_id)
        dataset2_id = self._run_random_lines1(history_id, dataset1_id)["outputs"][0]["id"]
        prov = self.gi.histories.show_dataset_provenance(history_id, dataset2_id)
        for key in ALL_PROV_KEYS:
            assert key in prov
        for key in MINIMAL_PROV_KEYS:
            assert key in prov["parameters"]["input"]
        for key in OTHER_PROV_KEYS:
            assert key not in prov["parameters"]["input"]
        recursive_prov = self.gi.histories.show_dataset_provenance(history_id, dataset2_id, follow=True)
        for key in ALL_PROV_KEYS:
            assert key in recursive_prov
        for key in ALL_PROV_KEYS:
            assert key in recursive_prov["parameters"]["input"]

    def test_delete_dataset(self):
        history_id = self.history["id"]
        dataset1_id = self._test_dataset(history_id)
        self.gi.histories.delete_dataset(history_id, dataset1_id)
        dataset = self.gi.histories.show_dataset(history_id, dataset1_id)
        assert dataset["deleted"]
        assert not dataset["purged"]

    def test_purge_dataset(self):
        history_id = self.history["id"]
        dataset1_id = self._test_dataset(history_id)
        self.gi.histories.delete_dataset(history_id, dataset1_id, purge=True, wait=True)
        dataset = self.gi.histories.show_dataset(history_id, dataset1_id)
        assert dataset["deleted"]
        assert dataset["purged"]

    def test_update_dataset(self):
        history_id = self.history["id"]
        dataset1_id = self._test_dataset(history_id)
        updated_dataset = self.gi.histories.update_dataset(history_id, dataset1_id, visible=False)
        if "id" not in updated_dataset:
            updated_dataset = self.gi.histories.show_dataset(history_id, dataset1_id)
        assert not updated_dataset["visible"]

    def test_upload_dataset_from_library(self):
        pass

    # download_dataset() is already tested in TestGalaxyDatasets

    def test_delete_history(self):
        result = self.gi.histories.delete_history(self.history["id"])
        assert result["deleted"]

        all_histories = self.gi.histories.get_histories()
        assert not any(d["id"] == self.history["id"] for d in all_histories)

    def test_undelete_history(self):
        self.gi.histories.delete_history(self.history["id"])
        self.gi.histories.undelete_history(self.history["id"])
        all_histories = self.gi.histories.get_histories()
        assert any(d["id"] == self.history["id"] for d in all_histories)

    def test_get_status(self):
        state = self.gi.histories.get_status(self.history["id"])
        assert "new" == state["state"]

    def test_get_most_recently_used_history(self):
        most_recently_used_history = self.gi.histories.get_most_recently_used_history()
        # if the user has been created via the API, it does not have
        # a session, therefore no history
        if most_recently_used_history is not None:
            assert most_recently_used_history["id"] is not None
            assert most_recently_used_history["name"] is not None
            assert most_recently_used_history["state"] is not None

    def test_download_history(self):
        jeha_id = self.gi.histories.export_history(self.history["id"], wait=True, maxwait=60)
        assert jeha_id
        tempdir = tempfile.mkdtemp(prefix="bioblend_test_")
        temp_fn = os.path.join(tempdir, "export.tar.gz")
        try:
            with open(temp_fn, "wb") as fo:
                self.gi.histories.download_history(self.history["id"], jeha_id, fo)
            assert tarfile.is_tarfile(temp_fn)
        finally:
            shutil.rmtree(tempdir)

    def test_import_history(self):
        path = test_util.get_abspath(os.path.join("data", "Galaxy-History-Test-history-for-export.tar.gz"))
        self.gi.histories.import_history(file_path=path)

    def test_copy_dataset(self):
        history_id = self.history["id"]
        contents = "1\t2\t3"
        dataset1_id = self._test_dataset(history_id, contents=contents)
        self.history_id2 = self.gi.histories.create_history("TestCopyDataset")["id"]
        copied_dataset = self.gi.histories.copy_dataset(self.history_id2, dataset1_id)
        expected_contents = ("\n".join(contents.splitlines()) + "\n").encode()
        self._wait_and_verify_dataset(copied_dataset["id"], expected_contents)
        self.gi.histories.delete_history(self.history_id2, purge=True)

    @test_util.skip_unless_galaxy("release_20.09")
    def test_update_dataset_datatype(self):
        history_id = self.history["id"]
        dataset1_id = self._test_dataset(history_id)
        self._wait_and_verify_dataset(dataset1_id, b"1\t2\t3\n")
        original_hda = self.gi.datasets.show_dataset(dataset1_id)
        assert original_hda["extension"] == "bed"
        self.gi.histories.update_dataset(history_id, dataset1_id, datatype="tabular")
        updated_hda = self.gi.datasets.show_dataset(dataset1_id)
        assert updated_hda["extension"] == "tabular"

    def test_get_extra_files(self):
        history_id = self.history["id"]
        dataset_id = self._test_dataset(history_id)
        extra_files = self.gi.histories.get_extra_files(history_id, dataset_id)
        assert extra_files == []

    def tearDown(self):
        self.gi.histories.delete_history(self.history["id"], purge=True)



# pylint: disable=W0622,E1101

"""
A basic object-oriented interface for Galaxy entities.
"""

import abc
import json
from collections.abc import (
    Mapping,
    Sequence,
)
from typing import (
    Any,
    Callable,
    cast,
    ClassVar,
    Dict,
    Generic,
    IO,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

import bioblend
from bioblend.galaxy.workflows import InputsBy
from bioblend.util import abstractclass

if TYPE_CHECKING:
    from . import client
    from .galaxy_instance import GalaxyInstance

__all__ = (
    "Wrapper",
    "Step",
    "Workflow",
    "LibraryContentInfo",
    "HistoryContentInfo",
    "DatasetContainer",
    "History",
    "Library",
    "Folder",
    "Dataset",
    "HistoryDatasetAssociation",
    "DatasetCollection",
    "HistoryDatasetCollectionAssociation",
    "LibraryDatasetDatasetAssociation",
    "LibraryDataset",
    "Tool",
    "Job",
    "LibraryPreview",
    "HistoryPreview",
    "WorkflowPreview",
)

WrapperSubtype = TypeVar("WrapperSubtype", bound="Wrapper")


@abstractclass
class Wrapper:
    """
    Abstract base class for Galaxy entity wrappers.

    Wrapper instances wrap deserialized JSON dictionaries such as the
    ones obtained by the Galaxy web API, converting key-based access to
    attribute-based access (e.g., ``library['name'] -> library.name``).

    Dict keys that are converted to attributes are listed in the
    ``BASE_ATTRS`` class variable: this is the 'stable' interface.
    Note that the wrapped dictionary is accessible via the ``wrapped``
    attribute.
    """

    BASE_ATTRS: Tuple[str, ...] = ("id",)
    gi: Optional["GalaxyInstance"]
    id: str
    is_modified: bool
    wrapped: dict
    _cached_parent: Optional["Wrapper"]

    def __init__(
        self, wrapped: Dict[str, Any], parent: Optional["Wrapper"] = None, gi: Optional["GalaxyInstance"] = None
    ) -> None:
        """
        :type wrapped: dict
        :param wrapped: JSON-serializable dictionary

        :type parent: :class:`Wrapper`
        :param parent: the parent of this wrapper

        :type gi: :class:`GalaxyInstance`
        :param gi: the GalaxyInstance through which we can access this wrapper
        """
        if not isinstance(wrapped, Mapping):
            raise TypeError("wrapped object must be a mapping type")
        # loads(dumps(x)) is a bit faster than deepcopy and allows type checks
        try:
            dumped = json.dumps(wrapped)
        except (TypeError, ValueError):
            raise ValueError("wrapped object must be JSON-serializable")
        object.__setattr__(self, "wrapped", json.loads(dumped))
        for k in self.BASE_ATTRS:
            object.__setattr__(self, k, self.wrapped.get(k))
        object.__setattr__(self, "_cached_parent", parent)
        object.__setattr__(self, "is_modified", False)
        object.__setattr__(self, "gi", gi)

    @property
    def parent(self) -> Optional["Wrapper"]:
        """
        The parent of this wrapper.
        """
        return self._cached_parent

    @property
    def is_mapped(self) -> bool:
        """
        ``True`` if this wrapper is mapped to an actual Galaxy entity.
        """
        return self.id is not None

    def unmap(self) -> None:
        """
        Disconnect this wrapper from Galaxy.
        """
        object.__setattr__(self, "id", None)

    def clone(self: WrapperSubtype) -> WrapperSubtype:
        """
        Return an independent copy of this wrapper.
        """
        return self.__class__(self.wrapped)

    def touch(self) -> None:
        """
        Mark this wrapper as having been modified since its creation.
        """
        object.__setattr__(self, "is_modified", True)
        if self.parent:
            self.parent.touch()

    def to_json(self) -> str:
        """
        Return a JSON dump of this wrapper.
        """
        return json.dumps(self.wrapped)

    @classmethod
    def from_json(cls: Type[WrapperSubtype], jdef: str) -> WrapperSubtype:
        """
        Build a new wrapper from a JSON dump.
        """
        return cls(json.loads(jdef))

    # FIXME: things like self.x[0] = 'y' do NOT call self.__setattr__
    def __setattr__(self, name: str, value: str) -> None:
        if name not in self.wrapped:
            raise AttributeError("can't set attribute")
        self.wrapped[name] = value
        object.__setattr__(self, name, value)
        self.touch()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.wrapped!r})"


class Step(Wrapper):
    """
    Workflow step.

    Steps are the main building blocks of a Galaxy workflow. A step can be: an
    input (type ``data_collection_input``, ``data_input`` or
    ``parameter_input``), a computational tool (type ``tool``), a subworkflow
    (type ``subworkflow``) or a pause (type ``pause``).
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "input_steps",
        "name",
        "tool_id",
        "tool_inputs",
        "tool_version",
        "type",
    )
    input_steps: Dict[str, Dict]
    type: str
    tool_id: Optional[str]
    tool_inputs: Dict
    tool_version: Optional[str]

    def __init__(self, step_dict: Dict[str, Any], parent: Wrapper) -> None:
        super().__init__(step_dict, parent=parent, gi=parent.gi)
        try:
            stype = step_dict["type"]
        except KeyError:
            raise ValueError("not a step dict")
        if stype not in {"data_collection_input", "data_input", "parameter_input", "pause", "subworkflow", "tool"}:
            raise ValueError(f"Unknown step type: {stype!r}")


class InvocationStep(Wrapper):
    """
    Invocation step.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "action",
        "job_id",
        "order_index",
        "state",
        "update_time",
        "workflow_step_id",
        "workflow_step_label",
        "workflow_step_uuid",
    )
    action: Optional[object]
    gi: "GalaxyInstance"
    job_id: str
    order_index: int
    state: str
    update_time: str
    workflow_step_id: str
    workflow_step_label: str
    workflow_step_uuid: str

    @property
    def parent(self) -> Wrapper:
        ret = super().parent
        assert ret is not None
        return ret

    def __init__(self, wrapped: Dict[str, Any], parent: Wrapper, gi: "GalaxyInstance") -> None:
        super().__init__(wrapped, parent, gi)

    def refresh(self) -> "InvocationStep":
        """
        Re-fetch the attributes pertaining to this object.

        :return: self
        """
        step_dict = self.gi.gi.invocations.show_invocation_step(self.parent.id, self.id)
        self.__init__(step_dict, parent=self.parent, gi=self.gi)  # type: ignore[misc]
        return self

    def get_outputs(self) -> Dict[str, "HistoryDatasetAssociation"]:
        """
        Get the output datasets of the invocation step

        :rtype: dict of `HistoryDatasetAssociation`
        :return: dictionary mapping output names to history datasets
        """
        if not hasattr(self, "outputs"):
            self.refresh()
        return {name: self.gi.datasets.get(out_dict["id"]) for name, out_dict in self.wrapped["outputs"].items()}

    def get_output_collections(self) -> Dict[str, "HistoryDatasetCollectionAssociation"]:
        """
        Get the output dataset collections of the invocation step

        :rtype: dict of `HistoryDatasetCollectionAssociation`
        :return: dictionary mapping output names to history dataset collections
        """
        if not hasattr(self, "output_collections"):
            self.refresh()
        return {
            name: self.gi.dataset_collections.get(out_coll_dict["id"])
            for name, out_coll_dict in self.wrapped["output_collections"].items()
        }


class Workflow(Wrapper):
    """
    Workflows represent ordered sequences of computations on Galaxy.

    A workflow defines a sequence of steps that produce one or more
    results from an input dataset.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "deleted",
        "inputs",
        "latest_workflow_uuid",
        "name",
        "owner",
        "published",
        "steps",
        "tags",
    )
    dag: Dict[str, Set[str]]
    deleted: bool
    input_labels_to_ids: Dict[str, Set[str]]
    inputs: Dict[str, Dict]
    inv_dag: Dict[str, Set[str]]
    missing_ids: List
    name: str
    owner: str
    POLLING_INTERVAL = 10  # for output state monitoring
    published: bool
    sink_ids: Set[str]
    source_ids: Set[str]
    steps: Dict[str, Step]
    tags: List[str]
    tool_labels_to_ids: Dict[str, Set[str]]

    def __init__(self, wf_dict: Dict[str, Any], gi: Optional["GalaxyInstance"] = None) -> None:
        super().__init__(wf_dict, gi=gi)
        if gi:
            tools_list_by_id = [t.id for t in gi.tools.get_previews()]
        else:
            tools_list_by_id = []
        missing_ids = []
        tool_labels_to_ids: Dict[str, Set[str]] = {}
        for k, v in self.steps.items():
            step_dict = cast(Dict[str, Any], v)
            # convert step ids to str for consistency with outer keys
            step_dict["id"] = str(step_dict["id"])
            for i in step_dict["input_steps"].values():
                i["source_step"] = str(i["source_step"])
            step = Step(step_dict, self)
            self.steps[k] = step
            if step.type == "tool":
                if not step.tool_inputs or step.tool_id not in tools_list_by_id:
                    missing_ids.append(k)
                assert step.tool_id
                tool_labels_to_ids.setdefault(step.tool_id, set()).add(step.id)
        input_labels_to_ids: Dict[str, Set[str]] = {}
        for id_, d in self.inputs.items():
            input_labels_to_ids.setdefault(d["label"], set()).add(id_)
        object.__setattr__(self, "input_labels_to_ids", input_labels_to_ids)
        object.__setattr__(self, "tool_labels_to_ids", tool_labels_to_ids)
        dag, inv_dag = self._get_dag()
        heads, tails = set(dag), set(inv_dag)
        object.__setattr__(self, "dag", dag)
        object.__setattr__(self, "inv_dag", inv_dag)
        object.__setattr__(self, "source_ids", heads - tails)
        assert (
            set(self.inputs) == self.data_collection_input_ids | self.data_input_ids | self.parameter_input_ids
        ), f"inputs is {self.inputs!r}, while data_collection_input_ids is {self.data_collection_input_ids!r}, data_input_ids is {self.data_input_ids!r} and parameter_input_ids is {self.parameter_input_ids!r}"
        object.__setattr__(self, "sink_ids", tails - heads)
        object.__setattr__(self, "missing_ids", missing_ids)

    def _get_dag(self) -> Tuple[Dict[str, Set[str]], Dict[str, Set[str]]]:
        """
        Return the workflow's DAG.

        For convenience, this method computes a 'direct' (step =>
        successors) and an 'inverse' (step => predecessors)
        representation of the same DAG.

        For instance, a workflow with a single tool *c*, two inputs
        *a, b* and three outputs *d, e, f* is represented by (direct)::

          {'a': {'c'}, 'b': {'c'}, 'c': {'d', 'e', 'f'}}

        and by (inverse)::

          {'c': {'a', 'b'}, 'd': {'c'}, 'e': {'c'}, 'f': {'c'}}
        """
        dag: Dict[str, Set[str]] = {}
        inv_dag: Dict[str, Set[str]] = {}
        for s in self.steps.values():
            assert isinstance(s, Step)
            for i in s.input_steps.values():
                head, tail = i["source_step"], s.id
                dag.setdefault(head, set()).add(tail)
                inv_dag.setdefault(tail, set()).add(head)
        return dag, inv_dag

    def sorted_step_ids(self) -> List[str]:
        """
        Return a topological sort of the workflow's DAG.
        """
        ids: List[str] = []
        source_ids = self.source_ids.copy()
        inv_dag = {k: v.copy() for k, v in self.inv_dag.items()}
        while source_ids:
            head = source_ids.pop()
            ids.append(head)
            for tail in self.dag.get(head, set()):
                incoming = inv_dag[tail]
                incoming.remove(head)
                if not incoming:
                    source_ids.add(tail)
        return ids

    @property
    def data_input_ids(self) -> Set[str]:
        """
        Return the ids of data input steps for this workflow.
        """
        return {id_ for id_, s in self.steps.items() if s.type == "data_input"}

    @property
    def data_collection_input_ids(self) -> Set[str]:
        """
        Return the ids of data collection input steps for this workflow.
        """
        return {id_ for id_, s in self.steps.items() if s.type == "data_collection_input"}

    @property
    def parameter_input_ids(self) -> Set[str]:
        """
        Return the ids of parameter input steps for this workflow.
        """
        return {id_ for id_, s in self.steps.items() if s.type == "parameter_input"}

    @property
    def tool_ids(self) -> Set[str]:
        """
        Return the ids of tool steps for this workflow.
        """
        return {id_ for id_, s in self.steps.items() if s.type == "tool"}

    @property
    def input_labels(self) -> Set[str]:
        """
        Return the labels of this workflow's input steps.
        """
        return set(self.input_labels_to_ids)

    @property
    def is_runnable(self) -> bool:
        """
        Return True if the workflow can be run on Galaxy.

        A workflow is considered runnable on a Galaxy instance if all
        of the tools it uses are installed in that instance.
        """
        return not self.missing_ids

    @staticmethod
    def _convert_input_map(input_map: Dict[str, Any]) -> Dict[str, Any]:
        """
        Convert ``input_map`` to the format required by the Galaxy web API.

        :type input_map: dict
        :param input_map: a mapping to datasets or dataset collections

        :rtype: dict
        :return: a mapping in the format required by the Galaxy web API.
        """
        ret = {}
        for key, value in input_map.items():
            if isinstance(
                value,
                (
                    HistoryDatasetAssociation,
                    HistoryDatasetCollectionAssociation,
                    LibraryDatasetDatasetAssociation,
                    LibraryDataset,
                ),
            ):
                ret[key] = {"id": value.id, "src": value.SRC}
            else:
                ret[key] = value
        return ret

    # I think we should deprecate this method - NS
    def preview(self) -> "WorkflowPreview":
        assert self.gi is not None
        try:
            return [_ for _ in self.gi.workflows.get_previews(published=True) if _.id == self.id][0]
        except IndexError:
            raise ValueError(f"no object for id {self.id}")

    def export(self) -> Dict[str, Any]:
        """
        Export a re-importable representation of the workflow.

        :rtype: dict
        :return: a JSON-serializable dump of the workflow
        """
        assert self.gi is not None
        return self.gi.gi.workflows.export_workflow_dict(self.id)

    def delete(self) -> None:
        """
        Delete this workflow.

        .. warning::
          Deleting a workflow is irreversible - all of the data from
          the workflow will be permanently deleted.
        """
        assert self.gi is not None
        self.gi.workflows.delete(id_=self.id)
        self.unmap()

    def invoke(
        self,
        inputs: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        history: Optional[Union[str, "History"]] = None,
        import_inputs_to_history: bool = False,
        replacement_params: Optional[Dict[str, Any]] = None,
        allow_tool_state_corrections: bool = True,
        inputs_by: Optional[InputsBy] = None,
        parameters_normalized: bool = False,
    ) -> "Invocation":
        """
        Invoke the workflow. This will cause a workflow to be scheduled
        and return an object describing the workflow invocation.

        :type inputs: dict
        :param inputs: A mapping of workflow inputs to datasets and dataset collections.
                       The datasets source can be a LibraryDatasetDatasetAssociation (``ldda``),
                       LibraryDataset (``ld``), HistoryDatasetAssociation (``hda``), or
                       HistoryDatasetCollectionAssociation (``hdca``).

                       The map must be in the following format:
                       ``{'<input_index>': dataset or collection}``
                       (e.g. ``{'2': HistoryDatasetAssociation()}``)

                       This map may also be indexed by the UUIDs of the workflow steps,
                       as indicated by the ``uuid`` property of steps returned from the
                       Galaxy API. Alternatively workflow steps may be addressed by
                       the label that can be set in the workflow editor. If using
                       uuid or label you need to also set the ``inputs_by`` parameter
                       to ``step_uuid`` or ``name``.

        :type params: dict
        :param params: A mapping of non-datasets tool parameters (see below)

        :type history: History or str
        :param history: The history in which to store the workflow output, or
          the name of a new history to create. If ``None``, a new 'Unnamed
          history' is created.

        :type import_inputs_to_history: bool
        :param import_inputs_to_history: If ``True``, used workflow inputs will
          be imported into the history. If ``False``, only workflow outputs will
          be visible in the given history.

        :type allow_tool_state_corrections: bool
        :param allow_tool_state_corrections: If True, allow Galaxy to fill in
          missing tool state when running workflows. This may be useful for
          workflows using tools that have changed over time or for workflows
          built outside of Galaxy with only a subset of inputs defined.

        :type replacement_params: dict
        :param replacement_params: pattern-based replacements for post-job
          actions (see below)

        :type inputs_by: str
        :param inputs_by: Determines how inputs are referenced. Can be
          "step_index|step_uuid" (default), "step_index", "step_id", "step_uuid", or "name".

        :type parameters_normalized: bool
        :param parameters_normalized: Whether Galaxy should normalize ``params``
          to ensure everything is referenced by a numeric step ID. Default is
          ``False``, but when setting ``params`` for a subworkflow, ``True`` is
          required.

        :rtype: Invocation
        :return: the workflow invocation

        The ``params`` dict should be specified as follows::

          {STEP_ID: PARAM_DICT, ...}

        where PARAM_DICT is::

          {PARAM_NAME: VALUE, ...}

        For backwards compatibility, the following (deprecated) format is
        also supported for ``params``::

          {TOOL_ID: PARAM_DICT, ...}

        in which case PARAM_DICT affects all steps with the given tool id.
        If both by-tool-id and by-step-id specifications are used, the
        latter takes precedence.

        Finally (again, for backwards compatibility), PARAM_DICT can also
        be specified as::

          {'param': PARAM_NAME, 'value': VALUE}

        Note that this format allows only one parameter to be set per step.

        For a ``repeat`` parameter, the names of the contained parameters needs
        to be specified as ``<repeat name>_<repeat index>|<param name>``, with
        the repeat index starting at 0. For example, if the tool XML contains::

          <repeat name="cutoff" title="Parameters used to filter cells" min="1">
              <param name="name" type="text" value="n_genes" label="Name of param...">
                  <option value="n_genes">n_genes</option>
                  <option value="n_counts">n_counts</option>
              </param>
              <param name="min" type="float" min="0" value="0" label="Min value"/>
          </repeat>

        then the PARAM_DICT should be something like::

          {...
           "cutoff_0|name": "n_genes",
           "cutoff_0|min": "2",
           "cutoff_1|name": "n_counts",
           "cutoff_1|min": "4",
           ...}

        At the time of this writing, it is not possible to change the number of
        times the contained parameters are repeated. Therefore, the parameter
        indexes can go from 0 to n-1, where n is the number of times the
        repeated element was added when the workflow was saved in the Galaxy UI.

        The ``replacement_params`` dict should map parameter names in
        post-job actions (PJAs) to their runtime values. For
        instance, if the final step has a PJA like the following::

          {'RenameDatasetActionout_file1': {'action_arguments': {'newname': '${output}'},
                                            'action_type': 'RenameDatasetAction',
                                            'output_name': 'out_file1'}}

        then the following renames the output dataset to 'foo'::

          replacement_params = {'output': 'foo'}

        see also `this email thread
        <http://lists.bx.psu.edu/pipermail/galaxy-dev/2011-September/006875.html>`_.

        .. warning::
          Historically, workflow invocation consumed a ``dataset_map``
          data structure that was indexed by unencoded workflow step IDs. These
          IDs would not be stable across Galaxy instances. The new ``inputs``
          property is instead indexed by either the ``order_index`` property
          (which is stable across workflow imports) or the step UUID which is
          also stable.
        """
        assert self.gi is not None
        if not self.is_mapped:
            raise RuntimeError("workflow is not mapped to a Galaxy object")
        if not self.is_runnable:
            missing_tools_str = ", ".join(f"{self.steps[step_id].tool_id}[{step_id}]" for step_id in self.missing_ids)
            raise RuntimeError(f"workflow has missing tools: {missing_tools_str}")

        inv_dict = self.gi.gi.workflows.invoke_workflow(
            workflow_id=self.id,
            inputs=self._convert_input_map(inputs or {}),
            params=params,
            history_id=history.id if isinstance(history, History) else None,
            history_name=history if isinstance(history, str) else None,
            import_inputs_to_history=import_inputs_to_history,
            replacement_params=replacement_params,
            allow_tool_state_corrections=allow_tool_state_corrections,
            inputs_by=inputs_by,
            parameters_normalized=parameters_normalized,
        )
        return self.gi.invocations.get(inv_dict["id"])


class Invocation(Wrapper):
    """
    Invocation of a workflow.
    This causes the steps of a workflow to be executed in sequential order.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "history_id",
        "state",
        "update_time",
        "uuid",
        "workflow_id",
    )
    gi: "GalaxyInstance"
    history_id: str
    state: str
    steps: List[InvocationStep]
    update_time: str
    uuid: str
    workflow_id: str

    def __init__(self, inv_dict: Dict[str, Any], gi: "GalaxyInstance") -> None:
        super().__init__(inv_dict, gi=gi)
        self.steps = [InvocationStep(step, parent=self, gi=gi) for step in inv_dict["steps"]]
        self.inputs = [{**v, "label": k} for k, v in inv_dict["inputs"].items()]

    def sorted_step_ids(self) -> List[str]:
        """
        Get the step IDs sorted based on this order index.

        :rtype: list of str
        :param: sorted step IDs
        """
        return [step.id for step in sorted(self.steps, key=lambda step: step.order_index)]

    def step_states(self) -> Set[str]:
        """
        Get the set of step states for this invocation.

        :rtype: set
        :param: step states
        """
        return {step.state for step in self.steps}

    def number_of_steps(self) -> int:
        """
        Get the number of steps for this invocation.

        :rtype: int
        :param: number of steps
        """
        return len(self.steps)

    def sorted_steps_by(
        self,
        indices: Optional[Iterable[int]] = None,
        states: Optional[Iterable[Union[str, None]]] = None,
        step_ids: Optional[Iterable[str]] = None,
    ) -> List[InvocationStep]:
        """
        Get steps for this invocation, or get a subset by specifying
        optional parameters for filtering.

        :type indices: list of int
        :param indices: return steps that have matching order_index

        :type states: list of str
        :param states: return steps that have matching states

        :type step_ids: list of str
        :param step_ids: return steps that have matching step_ids

        :rtype: list of InvocationStep
        :param: invocation steps
        """
        steps: Union[List[InvocationStep], filter] = self.steps
        if indices is not None:
            steps = filter(lambda step: step.order_index in indices, steps)
        if states is not None:
            steps = filter(lambda step: step.state in states, steps)
        if step_ids is not None:
            steps = filter(lambda step: step.id in step_ids, steps)
        return sorted(steps, key=lambda step: step.order_index)

    def cancel(self) -> None:
        """
        Cancel this invocation.

        .. note::
          On success, this method updates the Invocation object's internal variables.
        """
        inv_dict = self.gi.gi.invocations.cancel_invocation(self.id)
        self.__init__(inv_dict, gi=self.gi)  # type: ignore[misc]

    def refresh(self) -> "Invocation":
        """
        Re-fetch the attributes pertaining to this object.

        :return: self
        """
        inv_dict = self.gi.gi.invocations.show_invocation(self.id)
        self.__init__(inv_dict, gi=self.gi)  # type: ignore[misc]
        return self

    def run_step_actions(self, steps: List[InvocationStep], actions: List[object]) -> None:
        """
        Run actions for active steps of this invocation.

        :type steps: list of InvocationStep
        :param steps: list of steps to run actions on

        :type actions: list of objects
        :param actions: list of actions to run

        .. note::
          On success, this method updates the Invocation object's internal step variables.
        """
        if not len(steps) == len(actions):
            raise RuntimeError(
                f"Different number of ``steps`` ({len(steps)}) and ``actions`` ({len(actions)}) in ``{self}.run_step_actions()``"
            )
        step_dict_list = [
            self.gi.gi.invocations.run_invocation_step_action(self.id, step.id, action)
            for step, action in zip(steps, actions)
        ]
        for step, step_dict in zip(steps, step_dict_list):
            step.__init__(step_dict, parent=self, gi=self.gi)  # type: ignore[misc]

    def summary(self) -> Dict[str, Any]:
        """
        Get a summary for this invocation.

        :rtype: dict
        :param: invocation summary
        """
        return self.gi.gi.invocations.get_invocation_summary(self.id)

    def step_jobs_summary(self) -> List[Dict[str, Any]]:
        """
        Get a summary for this invocation's step jobs.

        :rtype: list of dicts
        :param: step job summaries
        """
        return self.gi.gi.invocations.get_invocation_step_jobs_summary(self.id)

    def report(self) -> Dict[str, Any]:
        """
        Get a dictionary containing a Markdown report for this invocation.

        :rtype: dict
        :param: invocation report
        """
        return self.gi.gi.invocations.get_invocation_report(self.id)

    def save_report_pdf(self, file_path: str, chunk_size: int = bioblend.CHUNK_SIZE) -> None:
        """
        Download a PDF report for this invocation.

        :type file_path: str
        :param file_path: path to save the report

        :type chunk_size: int
        :param chunk_size: chunk size in bytes for reading remote data
        """
        self.gi.gi.invocations.get_invocation_report_pdf(self.id, file_path, chunk_size)

    def biocompute_object(self) -> Dict[str, Any]:
        """
        Get a BioCompute object for this invocation.

        :rtype: dict
        :param: BioCompute object
        """
        return self.gi.gi.invocations.get_invocation_biocompute_object(self.id)

    def wait(self, maxwait: float = 12000, interval: float = 3, check: bool = True) -> None:
        """
        Wait for this invocation to reach a terminal state.

        :type maxwait: float
        :param maxwait: upper limit on waiting time

        :type interval: float
        :param interval: polling interval in secconds

        :type check: bool
        :param check: if ``true``, raise an error if the terminal state is not 'scheduled'

        .. note::
          On success, this method updates the Invocation object's internal variables.
        """
        inv_dict = self.gi.gi.invocations.wait_for_invocation(self.id, maxwait=maxwait, interval=interval, check=check)
        self.__init__(inv_dict, gi=self.gi)  # type: ignore[misc]


DatasetSubtype = TypeVar("DatasetSubtype", bound="Dataset")


class Dataset(Wrapper, metaclass=abc.ABCMeta):
    """
    Abstract base class for Galaxy datasets.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "data_type",
        "file_ext",
        "file_name",
        "file_size",
        "genome_build",
        "misc_info",
        "name",
        "state",
    )
    container: "DatasetContainer"
    genome_build: str
    gi: "GalaxyInstance"
    misc_info: str
    name: str
    POLLING_INTERVAL = 1  # for state monitoring
    state: str

    def __init__(self, ds_dict: Dict[str, Any], container: "DatasetContainer", gi: "GalaxyInstance") -> None:
        super().__init__(ds_dict, gi=gi)
        object.__setattr__(self, "container", container)

    @property
    @abc.abstractmethod
    def _stream_url(self) -> str:
        """
        Return the URL to stream this dataset.
        """

    def get_stream(self, chunk_size: int = bioblend.CHUNK_SIZE) -> Iterator[bytes]:
        """
        Open dataset for reading and return an iterator over its contents.

        :type chunk_size: int
        :param chunk_size: read this amount of bytes at a time
        """
        kwargs: Dict[str, Any] = {"stream": True}
        if isinstance(self, LibraryDataset):
            kwargs["params"] = {"ld_ids%5B%5D": self.id}
        r = self.gi.gi.make_get_request(self._stream_url, **kwargs)
        if isinstance(self, LibraryDataset) and r.status_code == 500:
            # compatibility with older Galaxy releases
            kwargs["params"] = {"ldda_ids%5B%5D": self.id}
            r = self.gi.gi.make_get_request(self._stream_url, **kwargs)
        r.raise_for_status()
        return r.iter_content(chunk_size)  # FIXME: client can't close r

    def peek(self, chunk_size: int = bioblend.CHUNK_SIZE) -> bytes:
        """
        Open dataset for reading and return the first chunk.

        See :meth:`.get_stream` for param info.
        """
        try:
            return next(self.get_stream(chunk_size=chunk_size))
        except StopIteration:
            return b""

    def download(self, file_object: IO[bytes], chunk_size: int = bioblend.CHUNK_SIZE) -> None:
        """
        Open dataset for reading and save its contents to ``file_object``.

        :type file_object: file
        :param file_object: output file object

        See :meth:`.get_stream` for info on other params.
        """
        for chunk in self.get_stream(chunk_size=chunk_size):
            file_object.write(chunk)

    def get_contents(self, chunk_size: int = bioblend.CHUNK_SIZE) -> bytes:
        """
        Open dataset for reading and return its **full** contents.

        See :meth:`.get_stream` for param info.
        """
        return b"".join(self.get_stream(chunk_size=chunk_size))

    def refresh(self: DatasetSubtype) -> DatasetSubtype:
        """
        Re-fetch the attributes pertaining to this object.

        :return: self
        """
        gi_client = getattr(self.gi.gi, self.container.API_MODULE)
        ds_dict = gi_client.show_dataset(self.container.id, self.id)
        self.__init__(ds_dict, self.container, self.gi)  # type: ignore[misc]
        return self

    def wait(self, polling_interval: float = POLLING_INTERVAL, break_on_error: bool = True) -> None:
        """
        Wait for this dataset to come out of the pending states.

        :type polling_interval: float
        :param polling_interval: polling interval in seconds

        :type break_on_error: bool
        :param break_on_error: if ``True``, raise a RuntimeError exception if
          the dataset ends in the 'error' state.

        .. warning::

          This is a blocking operation that can take a very long time. Also,
          note that this method does not return anything; however, this dataset
          is refreshed (possibly multiple times) during the execution.
        """
        self.gi._wait_datasets([self], polling_interval=polling_interval, break_on_error=break_on_error)


class HistoryDatasetAssociation(Dataset):
    """
    Maps to a Galaxy ``HistoryDatasetAssociation``.
    """

    BASE_ATTRS = Dataset.BASE_ATTRS + ("annotation", "deleted", "purged", "tags", "visible")
    SRC = "hda"
    annotation: str
    deleted: bool
    purged: bool

    @property
    def _stream_url(self) -> str:
        base_url = self.gi.gi.histories._make_url(module_id=self.container.id, contents=True)
        return f"{base_url}/{self.id}/display"

    def get_stream(self, chunk_size: int = bioblend.CHUNK_SIZE) -> Iterator[bytes]:
        """
        Open dataset for reading and return an iterator over its contents.

        :type chunk_size: int
        :param chunk_size: read this amount of bytes at a time
        """
        _, _, r = self.gi.gi.datasets._initiate_download(
            self.id,
            stream_content=True,
        )
        return r.iter_content(chunk_size)  # FIXME: client can't close r

    def update(self, **kwargs: Any) -> "HistoryDatasetAssociation":
        """
        Update this history dataset metadata. Some of the attributes that can be
        modified are documented below.

        :type name: str
        :param name: Replace history dataset name with the given string

        :type genome_build: str
        :param genome_build: Replace history dataset genome build (dbkey)

        :type annotation: str
        :param annotation: Replace history dataset annotation with given string

        :type deleted: bool
        :param deleted: Mark or unmark history dataset as deleted

        :type visible: bool
        :param visible: Mark or unmark history dataset as visible
        """
        res = self.gi.gi.histories.update_dataset(self.container.id, self.id, **kwargs)
        # Refresh also the history because the dataset may have been (un)deleted
        self.container.refresh()
        self.__init__(res, self.container, gi=self.gi)  # type: ignore[misc]
        return self

    def delete(self, purge: bool = False, wait: bool = False) -> None:
        """
        Delete this history dataset.

        :type purge: bool
        :param purge: if ``True``, also purge (permanently delete) the dataset

        :param wait: Whether to wait for the dataset to be purged.

        .. note::
          The ``purge`` option works only if the Galaxy instance has the
          ``allow_user_dataset_purge`` option set to ``true`` in the
          ``config/galaxy.yml`` configuration file.
        """
        self.gi.gi.histories.delete_dataset(self.container.id, self.id, purge=purge, wait=wait)
        self.container.refresh()
        self.refresh()


DatasetCollectionSubtype = TypeVar("DatasetCollectionSubtype", bound="DatasetCollection")


class DatasetCollection(Wrapper, metaclass=abc.ABCMeta):
    """
    Abstract base class for Galaxy dataset collections.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "collection_type",
        "deleted",
        "name",
        "state",
    )
    container: Union["DatasetCollection", "History"]
    API_MODULE = "dataset_collections"
    collection_type: str
    deleted: bool
    gi: "GalaxyInstance"

    def __init__(
        self, dsc_dict: Dict[str, Any], container: Union["DatasetCollection", "History"], gi: "GalaxyInstance"
    ) -> None:
        super().__init__(dsc_dict, gi=gi)
        object.__setattr__(self, "container", container)

    def refresh(self: DatasetCollectionSubtype) -> DatasetCollectionSubtype:
        """
        Re-fetch the attributes pertaining to this object.

        :return: self
        """
        gi_client = getattr(self.gi.gi, self.container.API_MODULE)
        dsc_dict = gi_client.show_dataset_collection(self.container.id, self.id)
        self.__init__(dsc_dict, self.container, self.gi)  # type: ignore[misc]
        return self

    @abc.abstractmethod
    def delete(self) -> None:
        """
        Delete this dataset collection.
        """


class HistoryDatasetCollectionAssociation(DatasetCollection):
    """
    Maps to a Galaxy ``HistoryDatasetCollectionAssociation``.
    """

    BASE_ATTRS = DatasetCollection.BASE_ATTRS + ("tags", "visible", "elements")
    SRC = "hdca"
    elements: List[Dict]

    def delete(self) -> None:
        self.gi.gi.histories.delete_dataset_collection(self.container.id, self.id)
        self.container.refresh()
        self.refresh()


@abstractclass
class LibRelatedDataset(Dataset):
    """
    Base class for LibraryDatasetDatasetAssociation and LibraryDataset classes.
    """

    @property
    def _stream_url(self) -> str:
        base_url = self.gi.gi.libraries._make_url()
        return f"{base_url}/datasets/download/uncompressed"


class LibraryDatasetDatasetAssociation(LibRelatedDataset):
    """
    Maps to a Galaxy ``LibraryDatasetDatasetAssociation``.
    """

    BASE_ATTRS = LibRelatedDataset.BASE_ATTRS + ("deleted",)
    SRC = "ldda"


class LibraryDataset(LibRelatedDataset):
    """
    Maps to a Galaxy ``LibraryDataset``.
    """

    SRC = "ld"
    file_name: str

    def delete(self, purged: bool = False) -> None:
        """
        Delete this library dataset.

        :type purged: bool
        :param purged: if ``True``, also purge (permanently delete) the dataset
        """
        self.gi.gi.libraries.delete_library_dataset(self.container.id, self.id, purged=purged)
        self.container.refresh()
        self.refresh()

    def update(self, **kwargs: Any) -> "LibraryDataset":
        """
        Update this library dataset metadata. Some of the attributes that can be
        modified are documented below.

        :type name: str
        :param name: Replace history dataset name with the given string

        :type genome_build: str
        :param genome_build: Replace history dataset genome build (dbkey)
        """
        res = self.gi.gi.libraries.update_library_dataset(self.id, **kwargs)
        self.container.refresh()
        self.__init__(res, self.container, gi=self.gi)  # type: ignore[misc]
        return self


@abstractclass
class ContentInfo(Wrapper):
    """
    Instances of this class wrap dictionaries obtained by getting
    ``/api/{histories,libraries}/<ID>/contents`` from Galaxy.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "name",
        "type",
    )


class LibraryContentInfo(ContentInfo):
    """
    Instances of this class wrap dictionaries obtained by getting
    ``/api/libraries/<ID>/contents`` from Galaxy.
    """


class HistoryContentInfo(ContentInfo):
    """
    Instances of this class wrap dictionaries obtained by getting
    ``/api/histories/<ID>/contents`` from Galaxy.
    """

    BASE_ATTRS = ContentInfo.BASE_ATTRS + ("deleted", "state", "visible")


DatasetContainerSubtype = TypeVar("DatasetContainerSubtype", bound="DatasetContainer")


class DatasetContainer(Wrapper, Generic[DatasetSubtype], metaclass=abc.ABCMeta):
    """
    Abstract base class for dataset containers (histories and libraries).
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "deleted",
        "name",
    )
    API_MODULE: str
    CONTENT_INFO_TYPE: Type[ContentInfo]
    DS_TYPE: ClassVar[Callable]
    content_infos: List[ContentInfo]
    deleted: bool
    gi: "GalaxyInstance"
    name: str
    obj_gi_client: "client.ObjDatasetContainerClient"

    def __init__(
        self,
        c_dict: Dict[str, Any],
        content_infos: Optional[List[ContentInfo]] = None,
        gi: Optional["GalaxyInstance"] = None,
    ) -> None:
        """
        :type content_infos: list of :class:`ContentInfo`
        :param content_infos: info objects for the container's contents
        """
        assert gi is not None
        super().__init__(c_dict, gi=gi)
        if content_infos is None:
            content_infos = []
        object.__setattr__(self, "content_infos", content_infos)
        object.__setattr__(self, "obj_gi_client", getattr(self.gi, self.API_MODULE))

    @property
    def dataset_ids(self) -> List[str]:
        """
        Return the ids of the contained datasets.
        """
        return [_.id for _ in self.content_infos if _.type == "file"]

    # I think we should deprecate this method - NS
    def preview(self) -> "DatasetContainerPreview":
        getf = self.obj_gi_client.get_previews
        # self.state could be stale: check both regular and deleted containers
        try:
            p = [_ for _ in getf() if _.id == self.id][0]
        except IndexError:
            try:
                p = [_ for _ in getf(deleted=True) if _.id == self.id][0]
            except IndexError:
                raise ValueError(f"no object for id {self.id}")
        return p

    def refresh(self: DatasetContainerSubtype) -> DatasetContainerSubtype:
        """
        Re-fetch the attributes pertaining to this object.

        :return: self
        """
        fresh = self.obj_gi_client.get(self.id)
        self.__init__(fresh.wrapped, content_infos=fresh.content_infos, gi=self.gi)  # type: ignore[misc]
        return self

    def get_dataset(self, ds_id: str) -> DatasetSubtype:
        """
        Retrieve the dataset corresponding to the given id.

        :type ds_id: str
        :param ds_id: dataset id

        :rtype: :class:`~.HistoryDatasetAssociation` or
          :class:`~.LibraryDataset`
        :return: the dataset corresponding to ``ds_id``
        """
        gi_client = getattr(self.gi.gi, self.API_MODULE)
        ds_dict = gi_client.show_dataset(self.id, ds_id)
        return self.DS_TYPE(ds_dict, self, gi=self.gi)

    def get_datasets(self, name: Optional[str] = None) -> List[DatasetSubtype]:
        """
        Get all datasets contained inside this dataset container.

        :type name: str
        :param name: return only datasets with this name

        :rtype: list of :class:`~.HistoryDatasetAssociation` or list of
          :class:`~.LibraryDataset`
        :return: datasets with the given name contained inside this
          container

        .. note::
          when filtering library datasets by name, specify their full
          paths starting from the library's root folder, e.g.,
          ``/seqdata/reads.fastq``.  Full paths are available through
          the ``content_infos`` attribute of
          :class:`~.Library` objects.
        """
        if name is None:
            ds_ids = self.dataset_ids
        else:
            ds_ids = [_.id for _ in self.content_infos if _.name == name]
        return [self.get_dataset(_) for _ in ds_ids]

    @abc.abstractmethod
    def delete(self) -> None:
        """
        Delete this dataset container.
        """


class History(DatasetContainer[HistoryDatasetAssociation]):
    """
    Maps to a Galaxy history.
    """

    BASE_ATTRS = DatasetContainer.BASE_ATTRS + (
        "annotation",
        "published",
        "state",
        "state_ids",
        "state_details",
        "tags",
    )
    DS_TYPE = HistoryDatasetAssociation
    DSC_TYPE = HistoryDatasetCollectionAssociation
    CONTENT_INFO_TYPE = HistoryContentInfo
    API_MODULE = "histories"
    annotation: str
    published: bool
    tags: List[str]

    def update(self, **kwargs: Any) -> "History":
        """
        Update history metadata information. Some of the attributes that can be
        modified are documented below.

        :type name: str
        :param name: Replace history name with the given string

        :type annotation: str
        :param annotation: Replace history annotation with the given string

        :type deleted: bool
        :param deleted: Mark or unmark history as deleted

        :type purged: bool
        :param purged: If True, mark history as purged (permanently deleted).

        :type published: bool
        :param published: Mark or unmark history as published

        :type importable: bool
        :param importable: Mark or unmark history as importable

        :type tags: list
        :param tags: Replace history tags with the given list
        """
        # TODO: wouldn't it be better if name and annotation were attributes?
        self.gi.gi.histories.update_history(self.id, **kwargs)
        self.refresh()
        return self

    def delete(self, purge: bool = False) -> None:
        """
        Delete this history.

        :type purge: bool
        :param purge: if ``True``, also purge (permanently delete) the history

        .. note::
          The ``purge`` option works only if the Galaxy instance has the
          ``allow_user_dataset_purge`` option set to ``true`` in the
          ``config/galaxy.yml`` configuration file.
        """
        self.gi.histories.delete(id_=self.id, purge=purge)
        self.refresh()
        self.unmap()

    def import_dataset(self, lds: LibraryDataset) -> HistoryDatasetAssociation:
        """
        Import a dataset into the history from a library.

        :type lds: :class:`~.LibraryDataset`
        :param lds: the library dataset to import

        :rtype: :class:`~.HistoryDatasetAssociation`
        :return: the imported history dataset
        """
        if not self.is_mapped:
            raise RuntimeError("history is not mapped to a Galaxy object")
        if not isinstance(lds, LibraryDataset):
            raise TypeError("lds is not a LibraryDataset")
        res = self.gi.gi.histories.upload_dataset_from_library(self.id, lds.id)
        if not isinstance(res, Mapping):
            raise RuntimeError(f"upload_dataset_from_library: unexpected reply: {res!r}")
        self.refresh()
        return self.get_dataset(res["id"])

    def upload_file(self, path: str, **kwargs: Any) -> HistoryDatasetAssociation:
        """
        Upload the file specified by ``path`` to this history.

        :type path: str
        :param path: path of the file to upload

        See :meth:`~bioblend.galaxy.tools.ToolClient.upload_file` for
        the optional parameters.

        :rtype: :class:`~.HistoryDatasetAssociation`
        :return: the uploaded dataset
        """
        out_dict = self.gi.gi.tools.upload_file(path, self.id, **kwargs)
        self.refresh()
        return self.get_dataset(out_dict["outputs"][0]["id"])

    upload_dataset = upload_file

    def upload_from_ftp(self, path: str, **kwargs: Any) -> HistoryDatasetAssociation:
        """
        Upload the file specified by ``path`` from the user's FTP directory to
        this history.

        :type path: str
        :param path: path of the file in the user's FTP directory

        See :meth:`~bioblend.galaxy.tools.ToolClient.upload_file` for
        the optional parameters.

        :rtype: :class:`~.HistoryDatasetAssociation`
        :return: the uploaded dataset
        """
        out_dict = self.gi.gi.tools.upload_from_ftp(path, self.id, **kwargs)
        self.refresh()
        return self.get_dataset(out_dict["outputs"][0]["id"])

    def paste_content(self, content: str, **kwargs: Any) -> HistoryDatasetAssociation:
        """
        Upload a string to a new dataset in this history.

        :type content: str
        :param content: content of the new dataset to upload

        See :meth:`~bioblend.galaxy.tools.ToolClient.upload_file` for
        the optional parameters (except file_name).

        :rtype: :class:`~.HistoryDatasetAssociation`
        :return: the uploaded dataset
        """
        out_dict = self.gi.gi.tools.paste_content(content, self.id, **kwargs)
        self.refresh()
        return self.get_dataset(out_dict["outputs"][0]["id"])

    def export(
        self,
        gzip: bool = True,
        include_hidden: bool = False,
        include_deleted: bool = False,
        wait: bool = False,
        maxwait: Optional[int] = None,
    ) -> str:
        """
        Start a job to create an export archive for this history.  See
        :meth:`~bioblend.galaxy.histories.HistoryClient.export_history`
        for parameter and return value info.
        """
        return self.gi.gi.histories.export_history(
            self.id,
            gzip=gzip,
            include_hidden=include_hidden,
            include_deleted=include_deleted,
            wait=wait,
            maxwait=maxwait,
        )

    def download(self, jeha_id: str, outf: IO[bytes], chunk_size: int = bioblend.CHUNK_SIZE) -> None:
        """
        Download an export archive for this history.  Use :meth:`export`
        to create an export and get the required ``jeha_id``.  See
        :meth:`~bioblend.galaxy.histories.HistoryClient.download_history`
        for parameter and return value info.
        """
        return self.gi.gi.histories.download_history(self.id, jeha_id, outf, chunk_size=chunk_size)

    def create_dataset_collection(
        self,
        collection_description: bioblend.galaxy.dataset_collections.CollectionDescription,
        copy_elements: bool = True,
    ) -> "HistoryDatasetCollectionAssociation":
        """
        Create a new dataset collection in the history by providing a collection description.

        :type collection_description: bioblend.galaxy.dataset_collections.CollectionDescription
        :param collection_description: a description of the dataset collection

        :type copy_elements: bool
        :param copy_elements: Whether to make a copy of the elements of the
          collection being created

        :rtype: :class:`~.HistoryDatasetCollectionAssociation`
        :return: the new dataset collection
        """
        dataset_collection = self.gi.gi.histories.create_dataset_collection(
            self.id, collection_description, copy_elements=copy_elements
        )
        self.refresh()
        return self.get_dataset_collection(dataset_collection["id"])

    def get_dataset_collection(self, dsc_id: str) -> "HistoryDatasetCollectionAssociation":
        """
        Retrieve the dataset collection corresponding to the given id.

        :type dsc_id: str
        :param dsc_id: dataset collection id

        :rtype: :class:`~.HistoryDatasetCollectionAssociation`
        :return: the dataset collection corresponding to ``dsc_id``
        """
        dsc_dict = self.gi.gi.histories.show_dataset_collection(self.id, dsc_id)
        return self.DSC_TYPE(dsc_dict, self, gi=self.gi)


class Library(DatasetContainer[LibraryDataset]):
    """
    Maps to a Galaxy library.
    """

    BASE_ATTRS = DatasetContainer.BASE_ATTRS + ("description", "synopsis")
    DS_TYPE = LibraryDataset
    CONTENT_INFO_TYPE = LibraryContentInfo
    API_MODULE = "libraries"
    description: str
    synopsis: str

    @property
    def folder_ids(self) -> List[str]:
        """
        Return the ids of the contained folders.
        """
        return [_.id for _ in self.content_infos if _.type == "folder"]

    def delete(self) -> None:
        """
        Delete this library.
        """
        self.gi.libraries.delete(id_=self.id)
        self.refresh()
        self.unmap()

    def _pre_upload(self, folder: Optional["Folder"]) -> Optional[str]:
        """
        Return the id of the given folder, after sanity checking.
        """
        if not self.is_mapped:
            raise RuntimeError("library is not mapped to a Galaxy object")
        return None if folder is None else folder.id

    def upload_data(self, data: str, folder: Optional["Folder"] = None, **kwargs: Any) -> LibraryDataset:
        """
        Upload data to this library.

        :type data: str
        :param data: dataset contents

        :type folder: :class:`~.Folder`
        :param folder: a folder object, or ``None`` to upload to the root folder

        :rtype: :class:`~.LibraryDataset`
        :return: the dataset object that represents the uploaded content

        Optional keyword arguments: ``file_type``, ``dbkey``.
        """
        fid = self._pre_upload(folder)
        res = self.gi.gi.libraries.upload_file_contents(self.id, data, folder_id=fid, **kwargs)
        self.refresh()
        return self.get_dataset(res[0]["id"])

    def upload_from_url(self, url: str, folder: Optional["Folder"] = None, **kwargs: Any) -> LibraryDataset:
        """
        Upload data to this library from the given URL.

        :type url: str
        :param url: URL from which data should be read

        See :meth:`.upload_data` for info on other params.
        """
        fid = self._pre_upload(folder)
        res = self.gi.gi.libraries.upload_file_from_url(self.id, url, folder_id=fid, **kwargs)
        self.refresh()
        return self.get_dataset(res[0]["id"])

    def upload_from_local(self, path: str, folder: Optional["Folder"] = None, **kwargs: Any) -> LibraryDataset:
        """
        Upload data to this library from a local file.

        :type path: str
        :param path: local file path from which data should be read

        See :meth:`.upload_data` for info on other params.
        """
        fid = self._pre_upload(folder)
        res = self.gi.gi.libraries.upload_file_from_local_path(self.id, path, folder_id=fid, **kwargs)
        self.refresh()
        return self.get_dataset(res[0]["id"])

    def upload_from_galaxy_fs(
        self,
        paths: Union[str, Iterable[str]],
        folder: Optional["Folder"] = None,
        link_data_only: Literal["copy_files", "link_to_files"] = "copy_files",
        **kwargs: Any,
    ) -> List[LibraryDataset]:
        """
        Upload data to this library from filesystem paths on the server.

        :type paths: str or :class:`~collections.abc.Iterable` of str
        :param paths: server-side file paths from which data should be read

        :type link_data_only: str
        :param link_data_only: either 'copy_files' (default) or
          'link_to_files'. Setting to 'link_to_files' symlinks instead of
          copying the files

        :rtype: list of :class:`~.LibraryDataset`
        :return: the dataset objects that represent the uploaded content

        See :meth:`.upload_data` for info on other params.

        .. note::
          This method works only if the Galaxy instance has the
          ``allow_path_paste`` option set to ``true`` in the
          ``config/galaxy.yml`` configuration file.
        """
        fid = self._pre_upload(folder)
        if isinstance(paths, str):
            paths = (paths,)
        paths = "\n".join(paths)
        res = self.gi.gi.libraries.upload_from_galaxy_filesystem(
            self.id, paths, folder_id=fid, link_data_only=link_data_only, **kwargs
        )
        if res is None:
            raise RuntimeError("upload_from_galaxy_filesystem: no reply")
        if not isinstance(res, Sequence):
            raise RuntimeError(f"upload_from_galaxy_filesystem: unexpected reply: {res!r}")
        self.refresh()
        return [self.get_dataset(ds_info["id"]) for ds_info in res]

    def copy_from_dataset(
        self, hda: HistoryDatasetAssociation, folder: Optional["Folder"] = None, message: str = ""
    ) -> LibraryDataset:
        """
        Copy a history dataset into this library.

        :type hda: :class:`~.HistoryDatasetAssociation`
        :param hda: history dataset to copy into the library

        See :meth:`.upload_data` for info on other params.
        """
        fid = self._pre_upload(folder)
        res = self.gi.gi.libraries.copy_from_dataset(self.id, hda.id, folder_id=fid, message=message)
        self.refresh()
        return self.get_dataset(res["library_dataset_id"])

    def create_folder(
        self, name: str, description: Optional[str] = None, base_folder: Optional["Folder"] = None
    ) -> "Folder":
        """
        Create a folder in this library.

        :type name: str
        :param name: folder name

        :type description: str
        :param description: optional folder description

        :type base_folder: :class:`~.Folder`
        :param base_folder: parent folder, or ``None`` to create in the root
          folder

        :rtype: :class:`~.Folder`
        :return: the folder just created
        """
        bfid = None if base_folder is None else base_folder.id
        res = self.gi.gi.libraries.create_folder(self.id, name, description=description, base_folder_id=bfid)
        self.refresh()
        return self.get_folder(res[0]["id"])

    def get_folder(self, f_id: str) -> "Folder":
        """
        Retrieve the folder corresponding to the given id.

        :rtype: :class:`~.Folder`
        :return: the folder corresponding to ``f_id``
        """
        f_dict = self.gi.gi.libraries.show_folder(self.id, f_id)
        return Folder(f_dict, self, gi=self.gi)

    @property
    def root_folder(self) -> "Folder":
        """
        The root folder of this library.

        :rtype: :class:`~.Folder`
        :return: the root folder of this library
        """
        return self.get_folder(self.gi.gi.libraries._get_root_folder_id(self.id))


class Folder(Wrapper):
    """
    Maps to a folder in a Galaxy library.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "deleted",
        "description",
        "item_count",
        "name",
    )
    container: Library
    description: str
    gi: "GalaxyInstance"
    name: str
    _cached_parent: Optional["Folder"]

    def __init__(self, f_dict: Dict[str, Any], container: Library, gi: "GalaxyInstance") -> None:
        super().__init__(f_dict, gi=gi)
        object.__setattr__(self, "container", container)

    @property
    def parent(self) -> Optional["Folder"]:
        """
        The parent folder of this folder. The parent of the root folder is
        ``None``.

        :rtype: :class:`~.Folder`
        :return: the parent of this folder
        """
        if self._cached_parent is None:
            object.__setattr__(self, "_cached_parent", self._get_parent())
        return self._cached_parent

    def _get_parent(self) -> Optional["Folder"]:
        """
        Return the parent folder of this folder.
        """
        parent_id = self.wrapped["parent_id"]
        if parent_id is None:
            return None
        return self.container.get_folder(parent_id)

    def refresh(self) -> "Folder":
        """
        Re-fetch the attributes pertaining to this object.

        :return: self
        """
        f_dict = self.gi.gi.libraries.show_folder(self.container.id, self.id)
        self.__init__(f_dict, self.container, gi=self.gi)  # type: ignore[misc]
        return self


class Tool(Wrapper):
    """
    Maps to a Galaxy tool.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "name",
        "version",
    )
    gi: "GalaxyInstance"
    POLLING_INTERVAL = 10  # for output state monitoring

    def run(
        self, inputs: Dict[str, Any], history: History, wait: bool = False, polling_interval: float = POLLING_INTERVAL
    ) -> List[HistoryDatasetAssociation]:
        """
        Execute this tool in the given history with inputs from dict
        ``inputs``.

        :type inputs: dict
        :param inputs: dictionary of input datasets and parameters for
          the tool (see below)

        :type history: :class:`History`
        :param history: the history where to execute the tool

        :type wait: bool
        :param wait: whether to wait while the returned datasets are
          in a pending state

        :type polling_interval: float
        :param polling_interval: polling interval in seconds

        :rtype: list of :class:`HistoryDatasetAssociation`
        :return: list of output datasets

        The ``inputs`` dict should contain input datasets and parameters
        in the (largely undocumented) format used by the Galaxy API.
        Some examples can be found in `Galaxy's API test suite
        <https://github.com/galaxyproject/galaxy/blob/dev/lib/galaxy_test/api/test_tools.py>`_.
        The value of an input dataset can also be a :class:`Dataset`
        object, which will be automatically converted to the needed
        format.
        """
        for k, v in inputs.items():
            if isinstance(v, Dataset):
                inputs[k] = {"src": v.SRC, "id": v.id}  # type: ignore
        out_dict = self.gi.gi.tools.run_tool(history.id, self.id, inputs)
        outputs = [history.get_dataset(_["id"]) for _ in out_dict["outputs"]]
        if wait:
            self.gi._wait_datasets(outputs, polling_interval=polling_interval)
        return outputs


class Job(Wrapper):
    """
    Maps to a Galaxy job.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + ("state",)


DatasetContainerPreviewSubtype = TypeVar("DatasetContainerPreviewSubtype", bound="DatasetContainerPreview")


@abstractclass
class DatasetContainerPreview(Wrapper):
    """
    Abstract base class for dataset container (history and library) 'previews'.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "deleted",
        "name",
    )
    deleted: bool
    name: str


class LibraryPreview(DatasetContainerPreview):
    """
    Models Galaxy library 'previews'.

    Instances of this class wrap dictionaries obtained by getting
    ``/api/libraries`` from Galaxy.
    """


class HistoryPreview(DatasetContainerPreview):
    """
    Models Galaxy history 'previews'.

    Instances of this class wrap dictionaries obtained by getting
    ``/api/histories`` from Galaxy.
    """

    BASE_ATTRS = DatasetContainerPreview.BASE_ATTRS + (
        "annotation",
        "published",
        "purged",
        "tags",
    )


class WorkflowPreview(Wrapper):
    """
    Models Galaxy workflow 'previews'.

    Instances of this class wrap dictionaries obtained by getting
    ``/api/workflows`` from Galaxy.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "deleted",
        "latest_workflow_uuid",
        "name",
        "number_of_steps",
        "owner",
        "published",
        "show_in_tool_panel",
        "tags",
    )
    deleted: bool
    name: str
    owner: str
    published: bool
    show_in_tool_panel: bool
    tags: List[str]


class InvocationPreview(Wrapper):
    """
    Models Galaxy invocation 'previews'.

    Instances of this class wrap dictionaries obtained by getting
    ``/api/invocations`` from Galaxy.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + (
        "history_id",
        "state",
        "update_time",
        "uuid",
        "workflow_id",
    )
    history_id: str
    state: str
    update_time: str
    uuid: str
    workflow_id: str


class JobPreview(Wrapper):
    """
    Models Galaxy job 'previews'.

    Instances of this class wrap dictionaries obtained by getting
    ``/api/jobs`` from Galaxy.
    """

    BASE_ATTRS = Wrapper.BASE_ATTRS + ("state",)



import bioblend.galaxy
from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyUsers(GalaxyTestBase.GalaxyTestBase):
    def test_get_users(self):
        users = self.gi.users.get_users()
        for user in users:
            assert user["id"] is not None
            assert user["email"] is not None

    def test_show_user(self):
        current_user = self.gi.users.get_current_user()
        user = self.gi.users.show_user(current_user["id"])
        assert user["id"] == current_user["id"]
        assert user["username"] == current_user["username"]
        assert user["email"] == current_user["email"]
        # The 2 following tests randomly fail

    #        assert user["nice_total_disk_usage"] == current_user["nice_total_disk_usage"]
    #        assert user["total_disk_usage"] == current_user["total_disk_usage"]

    @test_util.skip_unless_galaxy("release_19.09")  # for user purging
    def test_create_remote_user(self):
        # WARNING: only admins can create users!
        if not self.gi.config.get_config()["use_remote_user"]:
            self.skipTest("This Galaxy instance is not configured to use remote users")
        new_user_email = "newuser@example.org"
        user = self.gi.users.create_remote_user(new_user_email)
        assert user["email"] == new_user_email
        if self.gi.config.get_config()["allow_user_deletion"]:
            deleted_user = self.gi.users.delete_user(user["id"])
            assert deleted_user["email"] == new_user_email
            assert deleted_user["deleted"]
            assert not deleted_user["purged"]

            purged_user = self.gi.users.delete_user(user["id"], purge=True)
            # email is redacted when purging a user
            assert purged_user["email"] != new_user_email
            assert purged_user["deleted"]
            assert purged_user["purged"]

    @test_util.skip_unless_galaxy("release_19.09")  # for user purging
    def test_create_local_user(self):
        # WARNING: only admins can create users!
        if self.gi.config.get_config()["use_remote_user"]:
            self.skipTest("This Galaxy instance is not configured to use local users")
        new_username = test_util.random_string()
        new_user_email = f"{new_username}@example.org"
        password = test_util.random_string(20)
        new_user = self.gi.users.create_local_user(new_username, new_user_email, password)
        assert new_user["username"] == new_username
        assert new_user["email"] == new_user_email
        # test a BioBlend GalaxyInstance can be created using username+password
        user_gi = bioblend.galaxy.GalaxyInstance(url=self.gi.base_url, email=new_user_email, password=password)
        assert user_gi.users.get_current_user()["email"] == new_user_email
        # test deletion and purging
        if self.gi.config.get_config()["allow_user_deletion"]:
            deleted_user = self.gi.users.delete_user(new_user["id"])
            assert deleted_user["username"] == new_username
            assert deleted_user["email"] == new_user_email
            assert deleted_user["deleted"]
            assert not deleted_user["purged"]

            purged_user = self.gi.users.delete_user(new_user["id"], purge=True)
            # username and email are redacted when purging a user
            assert purged_user["username"] != new_username
            assert purged_user["email"] != new_user_email
            assert purged_user["deleted"]
            assert purged_user["purged"]

    def test_get_current_user(self):
        user = self.gi.users.get_current_user()
        assert user["id"] is not None
        assert user["username"] is not None
        assert user["email"] is not None
        assert user["nice_total_disk_usage"] is not None
        assert user["total_disk_usage"] is not None

    @test_util.skip_unless_galaxy("release_19.09")  # for user purging
    def test_update_user(self):
        # WARNING: only admins can create users!
        if self.gi.config.get_config()["use_remote_user"]:
            self.skipTest("This Galaxy instance is not configured to use local users")
        new_username = test_util.random_string()
        new_user = self.gi.users.create_local_user(
            new_username, f"{new_username}@example.org", test_util.random_string(20)
        )
        new_user_id = new_user["id"]
        updated_username = test_util.random_string()
        updated_user_email = f"{updated_username}@example.org"
        self.gi.users.update_user(new_user_id, username=updated_username, email=updated_user_email)
        updated_user = self.gi.users.show_user(new_user_id)
        assert updated_user["username"] == updated_username
        assert updated_user["email"] == updated_user_email

        # delete and purge user after test (if possile)
        if self.gi.config.get_config()["allow_user_deletion"]:
            self.gi.users.delete_user(new_user_id)
            purged_user = self.gi.users.delete_user(new_user_id, purge=True)
            assert purged_user["deleted"]
            assert purged_user["purged"]

    def test_get_user_apikey(self):
        # Test getting the API key of the current user, which surely has one
        user_id = self.gi.users.get_current_user()["id"]
        apikey = self.gi.users.get_user_apikey(user_id)
        assert apikey and apikey != "Not available."
        # Test getting the API key of a new user, which doesn't have one
        new_username = test_util.random_string()
        new_user_id = self.gi.users.create_local_user(
            new_username, f"{new_username}@example.org", test_util.random_string(20)
        )["id"]
        assert self.gi.users.get_user_apikey(new_user_id) == "Not available."

    @test_util.skip_unless_galaxy("release_21.01")
    def test_get_or_create_user_apikey(self):
        # Check that get_or_create_user_apikey() doesn't regenerate an existing API key
        user_id = self.gi.users.get_current_user()["id"]
        apikey = self.gi.users.get_user_apikey(user_id)
        assert self.gi.users.get_or_create_user_apikey(user_id) == apikey
        # Check that get_or_create_user_apikey() generates an API key for a new user
        new_username = test_util.random_string()
        new_user_id = self.gi.users.create_local_user(
            new_username, f"{new_username}@example.org", test_util.random_string(20)
        )["id"]
        new_apikey = self.gi.users.get_or_create_user_apikey(new_user_id)
        assert new_apikey and new_apikey != "Not available."

    def test_create_user_apikey(self):
        # Test creating an API key for a new user
        new_username = test_util.random_string()
        new_user_id = self.gi.users.create_local_user(
            new_username, f"{new_username}@example.org", test_util.random_string(20)
        )["id"]
        new_apikey = self.gi.users.create_user_apikey(new_user_id)
        assert new_apikey and new_apikey != "Not available."
        # Test regenerating an API key for a user that already has one
        regenerated_apikey = self.gi.users.create_user_apikey(new_user_id)
        assert regenerated_apikey and regenerated_apikey not in (new_apikey, "Not available.")



"""
WARNING: only admins can operate on groups!
"""

import uuid

from . import GalaxyTestBase


class TestGalaxyGroups(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.name = f"test_{uuid.uuid4().hex}"
        self.group = self.gi.groups.create_group(self.name)[0]

    def tearDown(self):
        # As of 2015/04/13, deleting a group is not possible through the API
        pass

    def test_create_group(self):
        assert self.group["name"] == self.name
        assert self.group["id"] is not None

    def test_get_groups(self):
        groups = self.gi.groups.get_groups()
        for group in groups:
            assert group["id"] is not None
            assert group["name"] is not None

    def test_show_group(self):
        group_data = self.gi.groups.show_group(self.group["id"])
        assert self.group["id"] == group_data["id"]
        assert self.group["name"] == group_data["name"]

    def test_get_group_users(self):
        group_users = self.gi.groups.get_group_users(self.group["id"])
        assert group_users == []

    def test_get_group_roles(self):
        group_roles = self.gi.groups.get_group_roles(self.group["id"])
        assert group_roles == []

    def test_update_group(self):
        new_name = f"test_{uuid.uuid4().hex}"
        new_users = [self.gi.users.get_current_user()["id"]]
        self.gi.groups.update_group(self.group["id"], new_name, user_ids=new_users)
        updated_group = self.gi.groups.show_group(self.group["id"])
        assert self.group["id"] == updated_group["id"]
        assert updated_group["name"] == new_name
        updated_group_users = [_["id"] for _ in self.gi.groups.get_group_users(self.group["id"])]
        assert set(updated_group_users) == set(new_users)
        updated_group_roles = [_["id"] for _ in self.gi.groups.get_group_roles(self.group["id"])]
        assert set(updated_group_roles) == set()

    def test_add_delete_group_user(self):
        new_user = self.gi.users.get_current_user()["id"]
        ret = self.gi.groups.add_group_user(self.group["id"], new_user)
        assert ret["id"] == new_user
        updated_group_users = [_["id"] for _ in self.gi.groups.get_group_users(self.group["id"])]
        assert new_user in updated_group_users
        self.gi.groups.delete_group_user(self.group["id"], new_user)
        updated_group_users = [_["id"] for _ in self.gi.groups.get_group_users(self.group["id"])]
        assert new_user not in updated_group_users



"""
Test functions in bioblend.galaxy.container_resolution
"""

from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyContainerResolution(GalaxyTestBase.GalaxyTestBase):
    @test_util.skip_unless_galaxy("release_22.05")
    def test_get_container_resolvers(self):
        container_resolvers = self.gi.container_resolution.get_container_resolvers()
        assert isinstance(container_resolvers, list)
        assert len(container_resolvers) > 0
        assert isinstance(container_resolvers[0], dict)
        assert container_resolvers[0]["model_class"] == "ExplicitContainerResolver"
        assert container_resolvers[0]["resolver_type"] == "explicit"
        assert container_resolvers[0]["can_uninstall_dependencies"] is False
        assert container_resolvers[0]["builds_on_resolution"] is False

    @test_util.skip_unless_galaxy("release_22.05")
    def test_show_container_resolver(self):
        container_resolver = self.gi.container_resolution.show_container_resolver(0)
        print(container_resolver)
        assert isinstance(container_resolver, dict)
        assert container_resolver["model_class"] == "ExplicitContainerResolver"
        assert container_resolver["resolver_type"] == "explicit"
        assert container_resolver["can_uninstall_dependencies"] is False
        assert container_resolver["builds_on_resolution"] is False

    @test_util.skip_unless_galaxy("release_22.05")
    def test_resolve(self):
        tool = self.gi.container_resolution.resolve(tool_id="CONVERTER_parquet_to_csv")
        print(tool)
        assert isinstance(tool, dict)

        tool_requirements_only = self.gi.container_resolution.resolve(
            tool_id="CONVERTER_parquet_to_csv", requirements_only=True
        )
        assert isinstance(tool_requirements_only, dict)

    @test_util.skip_unless_galaxy("release_22.05")
    def test_resolve_toolbox(self):
        toolbox = self.gi.container_resolution.resolve_toolbox()
        assert isinstance(toolbox, list)
        assert len(toolbox) > 0
        assert isinstance(toolbox[0], dict)

        toolbox_by_tool_ids = self.gi.container_resolution.resolve_toolbox(tool_ids=[toolbox[0]["tool_id"]])
        assert isinstance(toolbox_by_tool_ids, list)
        assert len(toolbox_by_tool_ids) == 1
        assert isinstance(toolbox_by_tool_ids[0], dict)

        toolbox_by_resolver_type = self.gi.container_resolution.resolve_toolbox(resolver_type="mulled")
        assert isinstance(toolbox_by_resolver_type, list)
        assert len(toolbox_by_resolver_type) > 0
        assert isinstance(toolbox_by_resolver_type[0], dict)
        assert len(toolbox) == len(toolbox_by_resolver_type)
        for tool in toolbox_by_resolver_type:
            print(tool)
            assert (
                tool["status"]["dependency_type"] is None
                or tool["status"]["container_resolver"]["resolver_type"] == "mulled"
            )

        toolbox_by_container_type = self.gi.container_resolution.resolve_toolbox(container_type="docker")
        assert isinstance(toolbox_by_container_type, list)
        assert len(toolbox_by_container_type) > 0
        assert isinstance(toolbox_by_container_type[0], dict)
        assert len(toolbox) == len(toolbox_by_container_type)
        for tool in toolbox_by_container_type:
            assert tool["status"]["dependency_type"] is None or tool["status"]["dependency_type"] == "docker"
            assert (
                tool["status"]["dependency_type"] is None or tool["status"]["container_description"]["type"] == "docker"
            )

        toolbox_requirements_only = self.gi.container_resolution.resolve_toolbox(requirements_only=True)
        assert isinstance(toolbox_requirements_only, list)
        assert len(toolbox_requirements_only) > 0
        assert isinstance(toolbox_requirements_only[0], dict)
        assert len(toolbox) == len(toolbox_requirements_only)

        # TODO unless containers are available this may fallback to conda by default?
        #      depending on Galaxy's config
        # toolbox_by_index = self.gi.container_resolution.resolve_toolbox(tool_ids=[toolbox[0]['tool_id']], index=0, install=True)
        # assert isinstance(toolbox_by_index, list)
        # assert len(toolbox_by_index) > 0
        # assert isinstance(toolbox_by_index[0], dict)

    # TODO unless containers are available this may fallback to conda by default?
    #      depending on Galaxy's config
    # def test_resolve_toolbox_with_install(self):
    #     toolbox = self.gi.container_resolution.resolve_toolbox_with_install(tool_ids=[])
    #     assert isinstance(toolbox, list)
    #     assert len(toolbox) == 0



import os
from datetime import (
    datetime,
    timedelta,
)
from operator import itemgetter
from typing import Literal

from bioblend.galaxy.tools.inputs import (
    dataset,
    inputs,
)
from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyJobs(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.history_id = self.gi.histories.create_history(name="TestGalaxyJobs")["id"]
        self.dataset_contents = "line 1\nline 2\rline 3\r\nline 4"
        self.dataset_id = self._test_dataset(self.history_id, contents=self.dataset_contents)

    def tearDown(self):
        self.gi.histories.delete_history(self.history_id, purge=True)

    @test_util.skip_unless_tool("cat1")
    def test_wait_for_job(self):
        tool_inputs = inputs().set("input1", dataset(self.dataset_id))
        tool_output = self.gi.tools.run_tool(history_id=self.history_id, tool_id="cat1", tool_inputs=tool_inputs)
        job_id = tool_output["jobs"][0]["id"]
        job = self.gi.jobs.wait_for_job(job_id)
        assert job["state"] == "ok"

    @test_util.skip_unless_tool("random_lines1")
    def test_get_jobs(self):
        self._run_tool()
        self._run_tool()

        jobs = self.gi.jobs.get_jobs(tool_id="random_lines1", history_id=self.history_id)
        assert len(jobs) == 2
        jobs = self.gi.jobs.get_jobs(history_id=self.history_id, state="failed")
        assert len(jobs) == 0
        yesterday = datetime.today() - timedelta(days=1)
        jobs = self.gi.jobs.get_jobs(date_range_max=yesterday.strftime("%Y-%m-%d"), history_id=self.history_id)
        assert len(jobs) == 0
        tomorrow = datetime.today() + timedelta(days=1)
        jobs = self.gi.jobs.get_jobs(date_range_min=tomorrow.strftime("%Y-%m-%d"))
        assert len(jobs) == 0
        jobs = self.gi.jobs.get_jobs(date_range_min=datetime.today().strftime("%Y-%m-%d"), history_id=self.history_id)
        assert len(jobs) == 3

    @test_util.skip_unless_galaxy("release_21.05")
    def test_get_jobs_with_filtering(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        workflow_id = self.gi.workflows.import_workflow_from_local_path(path)["id"]
        dataset = {"src": "hda", "id": self.dataset_id}
        invocation1 = self.gi.workflows.invoke_workflow(
            workflow_id,
            inputs={"Input 1": dataset, "Input 2": dataset},
            history_id=self.history_id,
            inputs_by="name",
        )
        invocation2 = self.gi.workflows.invoke_workflow(
            workflow_id,
            inputs={"Input 1": dataset, "Input 2": dataset},
            history_id=self.history_id,
            inputs_by="name",
        )
        self.gi.invocations.wait_for_invocation(invocation1["id"])
        self.gi.invocations.wait_for_invocation(invocation2["id"])

        all_jobs = self.gi.jobs.get_jobs(history_id=self.history_id, order_by="create_time")
        assert len(all_jobs) == 3
        job1_id = all_jobs[1]["id"]
        jobs = self.gi.jobs.get_jobs(history_id=self.history_id, limit=1, offset=1, order_by="create_time")
        assert len(jobs) == 1
        assert jobs[0]["id"] == job1_id
        jobs = self.gi.jobs.get_jobs(invocation_id=invocation1["id"])
        assert len(jobs) == 1
        job_id_inv = jobs[0]["id"]
        jobs = self.gi.jobs.get_jobs(workflow_id=workflow_id)
        assert len(jobs) == 2
        assert job_id_inv in [job["id"] for job in jobs]

    @test_util.skip_unless_galaxy("release_21.01")
    @test_util.skip_unless_tool("random_lines1")
    def test_run_and_rerun_random_lines(self):
        original_output = self._run_tool(input_format="21.01")
        original_job_id = original_output["jobs"][0]["id"]

        rerun_output = self.gi.jobs.rerun_job(original_job_id)
        original_output_content = self.gi.datasets.download_dataset(original_output["outputs"][0]["id"])
        rerun_output_content = self.gi.datasets.download_dataset(rerun_output["outputs"][0]["id"])
        assert rerun_output_content == original_output_content

    @test_util.skip_unless_galaxy("release_21.01")
    @test_util.skip_unless_tool("Show beginning1")
    def test_rerun_and_remap(self):
        path = test_util.get_abspath(os.path.join("data", "select_first.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        wf_inputs = {
            "0": {"src": "hda", "id": self.dataset_id},
            "1": "-1",
        }
        invocation_id = self.gi.workflows.invoke_workflow(wf["id"], inputs=wf_inputs, history_id=self.history_id)["id"]
        invocation = self.gi.invocations.wait_for_invocation(invocation_id)
        job_steps = [step for step in invocation["steps"] if step["job_id"]]
        job_steps.sort(key=itemgetter("order_index"))
        try:
            self.gi.jobs.wait_for_job(job_steps[0]["job_id"])
        except Exception:
            pass  # indicates the job failed as expected
        else:
            raise Exception("The job should have failed")

        history_contents = self.gi.histories.show_history(self.history_id, contents=True)
        assert len(history_contents) == 3
        assert history_contents[1]["state"] == "error"
        assert history_contents[2]["state"] == "paused"

        # resume the paused step job
        resumed_outputs = self.gi.jobs.resume_job(job_steps[-1]["job_id"])
        assert resumed_outputs[0]["name"] == "out_file1"
        # the following does not pass stably - the job goes back to paused too quickly
        # history_contents_resumed = self.gi.histories.show_history(self.history_id, contents=True)
        # assert history_contents_resumed[2]["state"] != "paused"

        # now rerun and remap with correct input param
        failed_job_id = self.gi.datasets.show_dataset(history_contents[1]["id"])["creating_job"]
        tool_inputs_update = {"lineNum": "1"}
        rerun_job = self.gi.jobs.rerun_job(failed_job_id, remap=True, tool_inputs_update=tool_inputs_update)
        new_job_id = rerun_job["jobs"][0]["id"]

        # Wait for the last dataset in the history to be unpaused and complete
        last_dataset = self.gi.histories.show_history(self.history_id, contents=True)[-1]
        last_job_id = self.gi.datasets.show_dataset(last_dataset["id"])["creating_job"]
        self.gi.jobs.wait_for_job(new_job_id)
        self.gi.jobs.resume_job(last_job_id)  # last_job can get stuck on paused - resume it in case
        self.gi.jobs.wait_for_job(last_job_id)
        assert last_dataset["hid"] == 3
        assert last_dataset["id"] == history_contents[2]["id"]
        self._wait_and_verify_dataset(last_dataset["id"], b"line 1\tline 1\n")

    @test_util.skip_unless_tool("random_lines1")
    def test_get_common_problems(self):
        job_id = self._run_tool()["jobs"][0]["id"]
        response = self.gi.jobs.get_common_problems(job_id)
        assert response == {"has_duplicate_inputs": False, "has_empty_inputs": True}

    @test_util.skip_unless_tool("random_lines1")
    def test_get_inputs(self):
        job_id = self._run_tool()["jobs"][0]["id"]
        response = self.gi.jobs.get_inputs(job_id)
        assert response == [{"name": "input", "dataset": {"src": "hda", "id": self.dataset_id}}]

    @test_util.skip_unless_tool("random_lines1")
    def test_get_outputs(self):
        output = self._run_tool()
        job_id, output_id = output["jobs"][0]["id"], output["outputs"][0]["id"]
        response = self.gi.jobs.get_outputs(job_id)
        assert response == [{"name": "out_file1", "dataset": {"src": "hda", "id": output_id}}]

    @test_util.skip_unless_galaxy("release_20.05")
    @test_util.skip_unless_tool("random_lines1")
    def test_get_destination_params(self):
        job_id = self._run_tool()["jobs"][0]["id"]
        # In Galaxy 20.05 and 20.09 we need to wait for the job, otherwise
        # `get_destination_params()` receives a 500 error code. Fixed upstream
        # in https://github.com/galaxyproject/galaxy/commit/3e7f03cd1f229b8c9421ade02002728a33e131d8
        self.gi.jobs.wait_for_job(job_id)
        response = self.gi.jobs.get_destination_params(job_id)
        assert "Runner" in response
        assert "Runner Job ID" in response
        assert "Handler" in response

    @test_util.skip_unless_tool("random_lines1")
    def test_search_jobs(self):
        job_id = self._run_tool()["jobs"][0]["id"]
        inputs = {
            "num_lines": "1",
            "input": {"src": "hda", "id": self.dataset_id},
            "seed_source|seed_source_selector": "set_seed",
            "seed_source|seed": "asdf",
        }
        response = self.gi.jobs.search_jobs("random_lines1", inputs)
        assert job_id in [job["id"] for job in response]

    @test_util.skip_unless_galaxy("release_20.01")
    @test_util.skip_unless_tool("random_lines1")
    def test_report_error(self):
        output = self._run_tool()
        job_id, output_id = output["jobs"][0]["id"], output["outputs"][0]["id"]
        response = self.gi.jobs.report_error(job_id, output_id, "Test error")
        # expected response when the Galaxy server does not have mail configured
        assert response == {
            "messages": [
                [
                    "An error occurred sending the report by email: Mail is not configured for this Galaxy instance",
                    "danger",
                ]
            ]
        }

    @test_util.skip_unless_galaxy("release_20.05")
    def test_show_job_lock(self):
        status = self.gi.jobs.show_job_lock()
        assert not status

    @test_util.skip_unless_galaxy("release_20.05")
    def test_update_job_lock(self):
        status = self.gi.jobs.update_job_lock(active=True)
        assert status
        status = self.gi.jobs.update_job_lock(active=False)
        assert not status

    def test_cancel_job(self):
        job_id = self._run_tool()["jobs"][0]["id"]
        self.gi.jobs.cancel_job(job_id)
        job = self.gi.jobs.wait_for_job(job_id, check=False)
        assert job["state"] in ("deleted", "deleting")

    def _run_tool(self, input_format: Literal["21.01", "legacy"] = "legacy") -> dict:
        return super()._run_random_lines1(self.history_id, self.dataset_id, input_format=input_format)



"""
This example creates a new user and prints her API key. It is also used to
initialize a Galaxy server in Continuous Integration testing of BioBlend.

Usage: python3 create_user_get_api_key.py <Galaxy_URL> <Galaxy_API_key> <new_username> <new_user_email> <new_password>
"""

import sys

import bioblend.galaxy

if len(sys.argv) != 6:
    print(
        "Usage: python3 create_user_get_api_key.py <Galaxy_URL> <Galaxy_API_key> <new_username> <new_user_email> <new_password>"
    )
    sys.exit(1)
galaxy_url = sys.argv[1]
galaxy_api_key = sys.argv[2]

# Initiating Galaxy connection
gi = bioblend.galaxy.GalaxyInstance(galaxy_url, key=galaxy_api_key)

# Create a new user and get a new API key for her
new_user = gi.users.create_local_user(sys.argv[3], sys.argv[4], sys.argv[5])
new_api_key = gi.users.create_user_apikey(new_user["id"])
print(new_api_key)



"""
Contains possible interactions with the Galaxy Data Libraries
"""

import logging
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
)

from bioblend import (
    NotReady,
    wait_on,
)
from bioblend.galaxy.client import Client
from bioblend.galaxy.datasets import TERMINAL_STATES
from bioblend.util import attach_file

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

LinkDataOnly = Literal["copy_files", "link_to_files"]

log = logging.getLogger(__name__)


class LibraryClient(Client):
    module = "libraries"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def create_library(
        self, name: str, description: Optional[str] = None, synopsis: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Create a data library with the properties defined in the arguments.

        :type name: str
        :param name: Name of the new data library

        :type description: str
        :param description: Optional data library description

        :type synopsis: str
        :param synopsis: Optional data library synopsis

        :rtype: dict
        :return: Details of the created library.
          For example::

            {'id': 'f740ab636b360a70',
             'name': 'Library from bioblend',
             'url': '/api/libraries/f740ab636b360a70'}
        """
        payload = {"name": name}
        if description:
            payload["description"] = description
        if synopsis:
            payload["synopsis"] = synopsis
        return self._post(payload)

    def delete_library(self, library_id: str) -> Dict[str, Any]:
        """
        Delete a data library.

        :type library_id: str
        :param library_id: Encoded data library ID identifying the library to be
          deleted

        :rtype: dict
        :return: Information about the deleted library

        .. warning::
          Deleting a data library is irreversible - all of the data from the
          library will be permanently deleted.
        """
        return self._delete(id=library_id)

    def _show_item(self, library_id: str, item_id: str) -> Dict[str, Any]:
        """
        Get details about a given library item.
        """
        url = "/".join((self._make_url(library_id, contents=True), item_id))
        return self._get(url=url)

    def delete_library_dataset(self, library_id: str, dataset_id: str, purged: bool = False) -> Dict[str, Any]:
        """
        Delete a library dataset in a data library.

        :type library_id: str
        :param library_id: library id where dataset is found in

        :type dataset_id: str
        :param dataset_id: id of the dataset to be deleted

        :type purged: bool
        :param purged: Indicate that the dataset should be purged (permanently
          deleted)

        :rtype: dict
        :return: A dictionary containing the dataset id and whether the dataset
          has been deleted.
          For example::

            {'deleted': True,
             'id': '60e680a037f41974'}
        """
        url = "/".join((self._make_url(library_id, contents=True), dataset_id))
        return self._delete(payload={"purged": purged}, url=url)

    def update_library_dataset(self, dataset_id: str, **kwargs: Any) -> Dict[str, Any]:
        """
        Update library dataset metadata. Some of the attributes that can be
        modified are documented below.

        :type dataset_id: str
        :param dataset_id: id of the dataset to be updated

        :type name: str
        :param name: Replace library dataset name with the given string

        :type misc_info: str
        :param misc_info: Replace library dataset misc_info with given string

        :type file_ext: str
        :param file_ext: Replace library dataset extension (must exist in the Galaxy registry)

        :type genome_build: str
        :param genome_build: Replace library dataset genome build (dbkey)

        :type tags: list
        :param tags: Replace library dataset tags with the given list

        :rtype: dict
        :return: details of the updated dataset
        """
        url = "/".join((self._make_url(), "datasets", dataset_id))
        return self._patch(payload=kwargs, url=url)

    def show_dataset(self, library_id: str, dataset_id: str) -> Dict[str, Any]:
        """
        Get details about a given library dataset. The required ``library_id``
        can be obtained from the datasets's library content details.

        :type library_id: str
        :param library_id: library id where dataset is found in

        :type dataset_id: str
        :param dataset_id: id of the dataset to be inspected

        :rtype: dict
        :return: A dictionary containing information about the dataset in the
          library
        """
        return self._show_item(library_id, dataset_id)

    def wait_for_dataset(
        self, library_id: str, dataset_id: str, maxwait: float = 12000, interval: float = 3
    ) -> Dict[str, Any]:
        """
        Wait until the library dataset state is terminal ('ok', 'empty',
        'error', 'discarded' or 'failed_metadata').

        :type library_id: str
        :param library_id: library id where dataset is found in

        :type dataset_id: str
        :param dataset_id: id of the dataset to wait for

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the dataset state to
          become terminal. After this time, a ``TimeoutException`` will be
          raised.

        :type interval: float
        :param interval: Time (in seconds) to wait between 2 consecutive checks.

        :rtype: dict
        :return: A dictionary containing information about the dataset in the
          library
        """

        def check_and_get_library_dataset() -> Dict[str, Any]:
            dataset = self.show_dataset(library_id, dataset_id)
            state = dataset["state"]
            if state in TERMINAL_STATES:
                return dataset
            raise NotReady(f"Dataset {dataset_id} in library {library_id} is in non-terminal state {state}")

        return wait_on(check_and_get_library_dataset, maxwait=maxwait, interval=interval)

    def show_folder(self, library_id: str, folder_id: str) -> Dict[str, Any]:
        """
        Get details about a given folder. The required ``folder_id`` can be
        obtained from the folder's library content details.

        :type library_id: str
        :param library_id: library id to inspect folders in

        :type folder_id: str
        :param folder_id: id of the folder to be inspected

        :rtype: dict
        :return: Information about the folder
        """
        return self._show_item(library_id, folder_id)

    def _get_root_folder_id(self, library_id: str) -> str:
        """
        Find the root folder (i.e. '/') of a library.

        :type library_id: str
        :param library_id: library id to find root of
        """
        library_dict = self.show_library(library_id=library_id)
        return library_dict["root_folder_id"]

    def create_folder(
        self, library_id: str, folder_name: str, description: Optional[str] = None, base_folder_id: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Create a folder in a library.

        :type library_id: str
        :param library_id: library id to use

        :type folder_name: str
        :param folder_name: name of the new folder in the data library

        :type description: str
        :param description: description of the new folder in the data library

        :type base_folder_id: str
        :param base_folder_id: id of the folder where to create the new folder.
          If not provided, the root folder will be used

        :rtype: list
        :return: List with a single dictionary containing information about the new folder
        """
        # Get root folder ID if no ID was provided
        if base_folder_id is None:
            base_folder_id = self._get_root_folder_id(library_id)
        # Compose the payload
        payload = {
            "name": folder_name,
            "folder_id": base_folder_id,
            "create_type": "folder",
        }
        if description is not None:
            payload["description"] = description
        return self._post(payload, id=library_id, contents=True)

    def get_folders(
        self, library_id: str, folder_id: Optional[str] = None, name: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Get all the folders in a library, or select a subset by specifying a
        folder name for filtering.

        :type library_id: str
        :param library_id: library id to use

        :type name: str
        :param name: Folder name to filter on. For ``name`` specify the full
                     path of the folder starting from the library's root
                     folder, e.g. ``/subfolder/subsubfolder``.

        :rtype: list
        :return: list of dicts each containing basic information about a folder

        .. versionchanged:: 1.1.1
           Using the deprecated ``folder_id`` parameter now raises a
           ``ValueError`` exception.
        """
        if folder_id is not None:
            raise ValueError(
                "The folder_id parameter has been removed, use the show_folder() method to view details of a folder for which you know the ID."
            )
        library_contents = self.show_library(library_id=library_id, contents=True)
        if name is not None:
            folders = [_ for _ in library_contents if _["type"] == "folder" and _["name"] == name]
        else:
            folders = [_ for _ in library_contents if _["type"] == "folder"]
        return folders

    def get_libraries(
        self, library_id: Optional[str] = None, name: Optional[str] = None, deleted: Optional[bool] = False
    ) -> List[Dict[str, Any]]:
        """
        Get all libraries, or select a subset by specifying optional arguments
        for filtering (e.g. a library name).

        :type name: str
        :param name: Library name to filter on.

        :type deleted: bool
        :param deleted: If ``False`` (the default), return only non-deleted
          libraries. If ``True``, return only deleted libraries. If ``None``,
          return both deleted and non-deleted libraries.

        :rtype: list
        :return: list of dicts each containing basic information about a library

        .. versionchanged:: 1.1.1
           Using the deprecated ``library_id`` parameter now raises a
           ``ValueError`` exception.
        """
        if library_id is not None:
            raise ValueError(
                "The library_id parameter has been removed, use the show_library() method to view details of a library for which you know the ID."
            )
        libraries = self._get(params={"deleted": deleted})
        if name is not None:
            libraries = [_ for _ in libraries if _["name"] == name]
        return libraries

    def show_library(self, library_id: str, contents: bool = False) -> Dict[str, Any]:
        """
        Get information about a library.

        :type library_id: str
        :param library_id: filter for library by library id

        :type contents: bool
        :param contents: whether to get contents of the library (rather
          than just the library details)

        :rtype: dict
        :return: details of the given library
        """
        return self._get(id=library_id, contents=contents)

    def _do_upload(self, library_id: str, **kwargs: Any) -> List[Dict[str, Any]]:
        """
        Set up the POST request and do the actual data upload to a data library.
        This method should not be called directly but instead refer to the
        methods specific for the desired type of data upload.
        """
        folder_id = kwargs.get("folder_id")
        if folder_id is None:
            folder_id = self._get_root_folder_id(library_id)
        files_attached = False
        # Compose the payload dict
        payload = {
            "folder_id": folder_id,
            "file_type": kwargs.get("file_type", "auto"),
            "dbkey": kwargs.get("dbkey", "?"),
            "create_type": "file",
            "tag_using_filenames": kwargs.get("tag_using_filenames", False),
            "preserve_dirs": kwargs.get("preserve_dirs", False),
        }
        if kwargs.get("roles"):
            payload["roles"] = kwargs["roles"]
        if kwargs.get("link_data_only") and kwargs["link_data_only"] != "copy_files":
            payload["link_data_only"] = "link_to_files"
        if kwargs.get("tags"):
            payload["tags"] = kwargs["tags"]
        # upload options
        if kwargs.get("file_url") is not None:
            payload["upload_option"] = "upload_file"
            payload["files_0|url_paste"] = kwargs["file_url"]
        elif kwargs.get("pasted_content") is not None:
            payload["upload_option"] = "upload_file"
            payload["files_0|url_paste"] = kwargs["pasted_content"]
        elif kwargs.get("server_dir") is not None:
            payload["upload_option"] = "upload_directory"
            payload["server_dir"] = kwargs["server_dir"]
        elif kwargs.get("file_local_path") is not None:
            payload["upload_option"] = "upload_file"
            payload["files_0|file_data"] = attach_file(kwargs["file_local_path"])
            files_attached = True
        elif kwargs.get("filesystem_paths") is not None:
            payload["upload_option"] = "upload_paths"
            payload["filesystem_paths"] = kwargs["filesystem_paths"]

        try:
            return self._post(payload, id=library_id, contents=True, files_attached=files_attached)
        finally:
            if payload.get("files_0|file_data") is not None:
                payload["files_0|file_data"].close()

    def upload_file_from_url(
        self,
        library_id: str,
        file_url: str,
        folder_id: Optional[str] = None,
        file_type: str = "auto",
        dbkey: str = "?",
        tags: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Upload a file to a library from a URL.

        :type library_id: str
        :param library_id: id of the library where to place the uploaded file

        :type file_url: str
        :param file_url: URL of the file to upload

        :type folder_id: str
        :param folder_id: id of the folder where to place the uploaded file.
          If not provided, the root folder will be used

        :type file_type: str
        :param file_type: Galaxy file format name

        :type dbkey: str
        :param dbkey: Dbkey

        :type tags: list
        :param tags: A list of tags to add to the datasets

        :rtype: list
        :return: List with a single dictionary containing information about the LDDA
        """
        return self._do_upload(
            library_id, file_url=file_url, folder_id=folder_id, file_type=file_type, dbkey=dbkey, tags=tags
        )

    def upload_file_contents(
        self,
        library_id: str,
        pasted_content: str,
        folder_id: Optional[str] = None,
        file_type: str = "auto",
        dbkey: str = "?",
        tags: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Upload pasted_content to a data library as a new file.

        :type library_id: str
        :param library_id: id of the library where to place the uploaded file

        :type pasted_content: str
        :param pasted_content: Content to upload into the library

        :type folder_id: str
        :param folder_id: id of the folder where to place the uploaded file.
          If not provided, the root folder will be used

        :type file_type: str
        :param file_type: Galaxy file format name

        :type dbkey: str
        :param dbkey: Dbkey

        :type tags: list
        :param tags: A list of tags to add to the datasets

        :rtype: list
        :return: List with a single dictionary containing information about the LDDA
        """
        return self._do_upload(
            library_id, pasted_content=pasted_content, folder_id=folder_id, file_type=file_type, dbkey=dbkey, tags=tags
        )

    def upload_file_from_local_path(
        self,
        library_id: str,
        file_local_path: str,
        folder_id: Optional[str] = None,
        file_type: str = "auto",
        dbkey: str = "?",
        tags: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Read local file contents from file_local_path and upload data to a
        library.

        :type library_id: str
        :param library_id: id of the library where to place the uploaded file

        :type file_local_path: str
        :param file_local_path: path of local file to upload

        :type folder_id: str
        :param folder_id: id of the folder where to place the uploaded file.
          If not provided, the root folder will be used

        :type file_type: str
        :param file_type: Galaxy file format name

        :type dbkey: str
        :param dbkey: Dbkey

        :type tags: list
        :param tags: A list of tags to add to the datasets

        :rtype: list
        :return: List with a single dictionary containing information about the LDDA
        """
        return self._do_upload(
            library_id,
            file_local_path=file_local_path,
            folder_id=folder_id,
            file_type=file_type,
            dbkey=dbkey,
            tags=tags,
        )

    def upload_file_from_server(
        self,
        library_id: str,
        server_dir: str,
        folder_id: Optional[str] = None,
        file_type: str = "auto",
        dbkey: str = "?",
        link_data_only: Optional[LinkDataOnly] = None,
        roles: str = "",
        preserve_dirs: bool = False,
        tag_using_filenames: bool = False,
        tags: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Upload all files in the specified subdirectory of the Galaxy library
        import directory to a library.

        :type library_id: str
        :param library_id: id of the library where to place the uploaded file

        :type server_dir: str
        :param server_dir: relative path of the subdirectory of
          ``library_import_dir`` to upload. All and only the files (i.e. no
          subdirectories) contained in the specified directory will be
          uploaded

        :type folder_id: str
        :param folder_id: id of the folder where to place the uploaded files.
          If not provided, the root folder will be used

        :type file_type: str
        :param file_type: Galaxy file format name

        :type dbkey: str
        :param dbkey: Dbkey

        :type link_data_only: str
        :param link_data_only: either 'copy_files' (default) or
          'link_to_files'. Setting to 'link_to_files' symlinks instead of
          copying the files

        :type roles: str
        :param roles: ???

        :type preserve_dirs: bool
        :param preserve_dirs: Indicate whether to preserve the directory structure when importing dir

        :type tag_using_filenames: bool
        :param tag_using_filenames: Indicate whether to generate dataset tags
          from filenames.

          .. versionchanged:: 0.14.0
            Changed the default from ``True`` to ``False``.

        :type tags: list
        :param tags: A list of tags to add to the datasets

        :rtype: list
        :return: List with a single dictionary containing information about the LDDA

        .. note::
          This method works only if the Galaxy instance has the
          ``library_import_dir`` option configured in the ``config/galaxy.yml``
          configuration file.
        """
        return self._do_upload(
            library_id,
            server_dir=server_dir,
            folder_id=folder_id,
            file_type=file_type,
            dbkey=dbkey,
            link_data_only=link_data_only,
            roles=roles,
            preserve_dirs=preserve_dirs,
            tag_using_filenames=tag_using_filenames,
            tags=tags,
        )

    def upload_from_galaxy_filesystem(
        self,
        library_id: str,
        filesystem_paths: str,
        folder_id: Optional[str] = None,
        file_type: str = "auto",
        dbkey: str = "?",
        link_data_only: Optional[LinkDataOnly] = None,
        roles: str = "",
        preserve_dirs: bool = False,
        tag_using_filenames: bool = False,
        tags: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Upload a set of files already present on the filesystem of the Galaxy
        server to a library.

        :type library_id: str
        :param library_id: id of the library where to place the uploaded file

        :type filesystem_paths: str
        :param filesystem_paths: file paths on the Galaxy server to upload to
          the library, one file per line

        :type folder_id: str
        :param folder_id: id of the folder where to place the uploaded files.
          If not provided, the root folder will be used

        :type file_type: str
        :param file_type: Galaxy file format name

        :type dbkey: str
        :param dbkey: Dbkey

        :type link_data_only: str
        :param link_data_only: either 'copy_files' (default) or
          'link_to_files'. Setting to 'link_to_files' symlinks instead of
          copying the files

        :type roles: str
        :param roles: ???

        :type preserve_dirs: bool
        :param preserve_dirs: Indicate whether to preserve the directory structure when importing dir

        :type tag_using_filenames: bool
        :param tag_using_filenames: Indicate whether to generate dataset tags
          from filenames.

          .. versionchanged:: 0.14.0
            Changed the default from ``True`` to ``False``.

        :type tags: list
        :param tags: A list of tags to add to the datasets

        :rtype: list
        :return: List of dictionaries containing information about each uploaded
          LDDA.

        .. note::
          This method works only if the Galaxy instance has the
          ``allow_path_paste`` option set to ``true`` in the
          ``config/galaxy.yml`` configuration file.
        """
        return self._do_upload(
            library_id,
            filesystem_paths=filesystem_paths,
            folder_id=folder_id,
            file_type=file_type,
            dbkey=dbkey,
            link_data_only=link_data_only,
            roles=roles,
            preserve_dirs=preserve_dirs,
            tag_using_filenames=tag_using_filenames,
            tags=tags,
        )

    def copy_from_dataset(
        self, library_id: str, dataset_id: str, folder_id: Optional[str] = None, message: str = ""
    ) -> Dict[str, Any]:
        """
        Copy a Galaxy dataset into a library.

        :type library_id: str
        :param library_id: id of the library where to place the uploaded file

        :type dataset_id: str
        :param dataset_id: id of the dataset to copy from

        :type folder_id: str
        :param folder_id: id of the folder where to place the uploaded files.
          If not provided, the root folder will be used

        :type message: str
        :param message: message for copying action

        :rtype: dict
        :return: LDDA information
        """
        if folder_id is None:
            folder_id = self._get_root_folder_id(library_id)
        payload = {
            "folder_id": folder_id,
            "create_type": "file",
            "from_hda_id": dataset_id,
            "ldda_message": message,
        }
        return self._post(payload, id=library_id, contents=True)

    def get_library_permissions(self, library_id: str) -> Dict[str, Any]:
        """
        Get the permissions for a library.

        :type library_id: str
        :param library_id: id of the library

        :rtype: dict
        :return: dictionary with all applicable permissions' values
        """
        url = self._make_url(library_id) + "/permissions"
        return self._get(url=url)

    def get_dataset_permissions(self, dataset_id: str) -> Dict[str, Any]:
        """
        Get the permissions for a dataset.

        :type dataset_id: str
        :param dataset_id: id of the dataset

        :rtype: dict
        :return: dictionary with all applicable permissions' values
        """
        url = "/".join((self._make_url(), "datasets", dataset_id, "permissions"))
        return self._get(url=url)

    def set_library_permissions(
        self,
        library_id: str,
        access_in: Optional[List[str]] = None,
        modify_in: Optional[List[str]] = None,
        add_in: Optional[List[str]] = None,
        manage_in: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Set the permissions for a library. Note: it will override all security
        for this library even if you leave out a permission type.

        :type library_id: str
        :param library_id: id of the library

        :type access_in: list
        :param access_in: list of role ids

        :type modify_in: list
        :param modify_in: list of role ids

        :type add_in: list
        :param add_in: list of role ids

        :type manage_in: list
        :param manage_in: list of role ids

        :rtype: dict
        :return: General information about the library
        """
        payload: Dict[str, List[str]] = {}
        if access_in:
            payload["LIBRARY_ACCESS_in"] = access_in
        if modify_in:
            payload["LIBRARY_MODIFY_in"] = modify_in
        if add_in:
            payload["LIBRARY_ADD_in"] = add_in
        if manage_in:
            payload["LIBRARY_MANAGE_in"] = manage_in
        url = self._make_url(library_id) + "/permissions"
        return self._post(payload, url=url)

    def set_dataset_permissions(
        self,
        dataset_id: str,
        access_in: Optional[List[str]] = None,
        modify_in: Optional[List[str]] = None,
        manage_in: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Set the permissions for a dataset. Note: it will override all security
        for this dataset even if you leave out a permission type.

        :type dataset_id: str
        :param dataset_id: id of the dataset

        :type access_in: list
        :param access_in: list of role ids

        :type modify_in: list
        :param modify_in: list of role ids

        :type manage_in: list
        :param manage_in: list of role ids

        :rtype: dict
        :return: dictionary with all applicable permissions' values
        """
        # we need here to define an action
        payload: Dict[str, Any] = {
            "action": "set_permissions",
        }
        if access_in:
            payload["access_ids[]"] = access_in
        if modify_in:
            payload["modify_ids[]"] = modify_in
        if manage_in:
            payload["manage_ids[]"] = manage_in
        url = "/".join((self._make_url(), "datasets", dataset_id, "permissions"))
        return self._post(payload, url=url)



import uuid

from . import GalaxyTestBase


class TestGalaxyRoles(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.name = f"test_{uuid.uuid4().hex}"
        self.description = "automated test role"
        self.role = self.gi.roles.create_role(self.name, self.description)

    def tearDown(self):
        # As of 2017/07/26, deleting a role is not possible through the API
        pass

    def test_get_roles(self):
        roles = self.gi.roles.get_roles()
        for role in roles:
            assert role["id"] is not None
            assert role["name"] is not None

    def test_create_role(self):
        assert self.role["name"] == self.name
        assert self.role["description"] == self.description
        assert self.role["id"] is not None



#!/usr/bin/env python
"""Wrapper around pytest to execute the bioblend Galaxy test suite against fixed instance.

By default all Galaxy tests will run but a smaller subset can be executed by setting
the environment variable ``BIOBLEND_TEST_SUITE`` to ``quick``.
"""
import os
import sys
from typing import (
    List,
    NoReturn,
    Optional,
)

try:
    import pytest
except ImportError:
    pytest = None

DIRECTORY = os.path.abspath(os.path.dirname(__file__))
BIOBLEND_TEST_SUITE = os.environ.get("BIOBLEND_TEST_SUITE", "full")

quick_tests = [
    "TestGalaxyRoles.py",
    "TestGalaxyRoles.py",
    "TestGalaxyUsers.py",
    "TestGalaxyToolData.py",
    "TestGalaxyTools.py::TestGalaxyTools::test_get_tools",  # Test single upload command.
]


def main(args: Optional[List[str]] = None) -> NoReturn:
    """Entry point that delegates to pytest.main."""
    if pytest is None:
        raise Exception("pytest is required to use this script.")
    if args is None:
        args = sys.argv[1:]
    if len(args) < 2:
        if BIOBLEND_TEST_SUITE == "full":
            args.append(os.path.join(DIRECTORY))
        else:
            for quick_test in quick_tests:
                args.append(os.path.join(DIRECTORY, quick_test))
    sys.exit(pytest.main(args))


if __name__ == "__main__":
    main()



"""
Contains possible interactions with the Galaxy Quota
"""

from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

QuotaOperations = Literal["+", "-", "="]


class QuotaClient(Client):
    module = "quotas"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_quotas(self, deleted: bool = False) -> List[Dict[str, Any]]:
        """
        Get a list of quotas

        :type deleted: bool
        :param deleted: Only return quota(s) that have been deleted

        :rtype: list
        :return: A list of dicts with details on individual quotas.
          For example::

            [{'id': '0604c8a56abe9a50',
              'model_class': 'Quota',
              'name': 'test ',
              'url': '/api/quotas/0604c8a56abe9a50'},
             {'id': '1ee267091d0190af',
              'model_class': 'Quota',
              'name': 'workshop',
              'url': '/api/quotas/1ee267091d0190af'}]
        """
        return self._get(deleted=deleted)

    def show_quota(self, quota_id: str, deleted: bool = False) -> Dict[str, Any]:
        """
        Display information on a quota

        :type quota_id: str
        :param quota_id: Encoded quota ID

        :type deleted: bool
        :param deleted: Search for quota in list of ones already marked as deleted

        :rtype: dict
        :return: A description of quota.
          For example::

            {'bytes': 107374182400,
             'default': [],
             'description': 'just testing',
             'display_amount': '100.0 GB',
             'groups': [],
             'id': '0604c8a56abe9a50',
             'model_class': 'Quota',
             'name': 'test ',
             'operation': '=',
             'users': []}
        """
        return self._get(id=quota_id, deleted=deleted)

    def create_quota(
        self,
        name: str,
        description: str,
        amount: str,
        operation: QuotaOperations,
        default: Optional[Literal["no", "registered", "unregistered"]] = "no",
        in_users: Optional[List[str]] = None,
        in_groups: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Create a new quota

        :type name: str
        :param name: Name for the new quota. This must be unique within a Galaxy instance.

        :type description: str
        :param description: Quota description

        :type amount: str
        :param amount: Quota size (E.g. ``10000MB``, ``99 gb``, ``0.2T``, ``unlimited``)

        :type operation: str
        :param operation: One of (``+``, ``-``, ``=``)

        :type default: str
        :param default: Whether or not this is a default quota. Valid values
                        are ``no``, ``unregistered``, ``registered``. None is
                        equivalent to ``no``.

        :type in_users: list of str
        :param in_users: A list of user IDs or user emails.

        :type in_groups: list of str
        :param in_groups: A list of group IDs or names.

        :rtype: dict
        :return: A description of quota.
          For example::

            {'url': '/galaxy/api/quotas/386f14984287a0f7',
             'model_class': 'Quota',
             'message': "Quota 'Testing' has been created with 1 associated users and 0 associated groups.",
             'id': '386f14984287a0f7',
             'name': 'Testing'}
        """
        payload: Dict[str, Any] = {
            "name": name,
            "description": description,
            "amount": amount,
            "operation": operation,
            "default": default,
        }
        if in_users:
            payload["in_users"] = in_users

        if in_groups:
            payload["in_groups"] = in_groups

        return self._post(payload)

    def update_quota(
        self,
        quota_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        amount: Optional[str] = None,
        operation: Optional[QuotaOperations] = None,
        default: str = "no",
        in_users: Optional[List[str]] = None,
        in_groups: Optional[List[str]] = None,
    ) -> str:
        """
        Update an existing quota

        :type quota_id: str
        :param quota_id: Encoded quota ID

        :type name: str
        :param name: Name for the new quota. This must be unique within a Galaxy instance.

        :type description: str
        :param description: Quota description. If you supply this parameter,
                            but not the name, an error will be thrown.

        :type amount: str
        :param amount: Quota size (E.g. ``10000MB``, ``99 gb``, ``0.2T``, ``unlimited``)

        :type operation: str
        :param operation: One of (``+``, ``-``, ``=``). If you wish to change this
                          value, you must also provide the ``amount``,
                          otherwise it will not take effect.

        :type default: str
        :param default: Whether or not this is a default quota. Valid values
                        are ``no``, ``unregistered``, ``registered``.
                        Calling this method with ``default="no"`` on a
                        non-default quota will throw an error. Not
                        passing this parameter is equivalent to passing ``no``.

        :type in_users: list of str
        :param in_users: A list of user IDs or user emails.

        :type in_groups: list of str
        :param in_groups: A list of group IDs or names.

        :rtype: str
        :return: A semicolon separated list of changes to the quota.
          For example::

            "Quota 'Testing-A' has been renamed to 'Testing-B'; Quota 'Testing-e' is now '-100.0 GB'; Quota 'Testing-B' is now the default for unregistered users"
        """
        payload: Dict[str, Any] = {"default": default}
        if name:
            payload["name"] = name

        if description:
            payload["description"] = description

        if amount:
            payload["amount"] = amount

        if operation:
            payload["operation"] = operation

        if in_users:
            payload["in_users"] = in_users

        if in_groups:
            payload["in_groups"] = in_groups

        return self._put(id=quota_id, payload=payload)

    def delete_quota(self, quota_id: str) -> str:
        """
        Delete a quota

        Before a quota can be deleted, the quota must not be a default quota.

        :type quota_id: str
        :param quota_id: Encoded quota ID.

        :rtype: str
        :return: A description of the changes, mentioning the deleted quota.
          For example::

            "Deleted 1 quotas: Testing-B"
        """
        return self._delete(id=quota_id)

    def undelete_quota(self, quota_id: str) -> str:
        """
        Undelete a quota

        :type quota_id: str
        :param quota_id: Encoded quota ID.

        :rtype: str
        :return: A description of the changes, mentioning the undeleted quota.
          For example::

            "Undeleted 1 quotas: Testing-B"
        """
        url = self._make_url(quota_id, deleted=True) + "/undelete"
        return self._post(url=url, payload={"id": quota_id})



"""
Contains possible interactions with the Galaxy Datatype
"""

from typing import (
    Dict,
    List,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class DatatypesClient(Client):
    module = "datatypes"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_datatypes(self, extension_only: bool = False, upload_only: bool = False) -> List[str]:
        """
        Get the list of all installed datatypes.

        :type extension_only: bool
        :param extension_only: Return only the extension rather than the datatype name

        :type upload_only: bool
        :param upload_only: Whether to return only datatypes which can be uploaded

        :rtype: list
        :return: A list of datatype names.
          For example::

            ['snpmatrix',
             'snptest',
             'tabular',
             'taxonomy',
             'twobit',
             'txt',
             'vcf',
             'wig',
             'xgmml',
             'xml']
        """

        params: Dict[str, bool] = {}
        if extension_only:
            params["extension_only"] = True

        if upload_only:
            params["upload_only"] = True

        return self._get(params=params)

    def get_sniffers(self) -> List[str]:
        """
        Get the list of all installed sniffers.

        :rtype: list
        :return: A list of sniffer names.
          For example::

            ['galaxy.datatypes.tabular:Vcf',
             'galaxy.datatypes.binary:TwoBit',
             'galaxy.datatypes.binary:Bam',
             'galaxy.datatypes.binary:Sff',
             'galaxy.datatypes.xml:Phyloxml',
             'galaxy.datatypes.xml:GenericXml',
             'galaxy.datatypes.sequence:Maf',
             'galaxy.datatypes.sequence:Lav',
             'galaxy.datatypes.sequence:csFasta']
        """
        url = self._make_url() + "/sniffers"
        return self._get(url=url)
